from quillsql.assets.pgtypes import PG_TYPES


def convert_type_to_postgres(data_type_id):
    # find the object in PG_TYPES that matches the type
    pg_type = next(
        (pg_type for pg_type in PG_TYPES if pg_type["oid"] == data_type_id), None
    )
    return pg_type["typname"] if pg_type else data_type_id



def remove_fields(query_result, fields_to_remove):
    if not isinstance(query_result, dict):
        return query_result
    fields = [
        {"name": field["name"], "dataTypeID": field["dataTypeID"]}
        for field in (query_result.get("fields") or [])
        if field.get("name") not in fields_to_remove
    ]
    rows = []
    for row in query_result.get("rows") or []:
        if not isinstance(row, dict):
            rows.append(row)
            continue
        filtered = dict(row)
        for field in fields_to_remove:
            if field in filtered:
                del filtered[field]
        rows.append(filtered)
    return {"fields": fields, "rows": rows}


def array_to_map(queries, array_to_map, metadata, target_pool):
    mapped_array = []
    for i in range(len(queries)):
        query_result = target_pool.query(queries[i])
        if isinstance(query_result, dict):
            mapped_array.append(query_result.get("rows"))
        else:
            mapped_array.append([])
    return mapped_array



from quillsql.db.postgres import (
    format_postgres,
    connect_to_postgres,
    connect_to_postgres_async,
    get_schema_column_info_postgres,
    get_tables_by_schema_postgres,
    run_query_postgres,
    disconnect_from_postgres,
    run_query_postgres_async,
    disconnect_from_postgres_async,
)
from quillsql.db.bigquery import (
    format_bigquery_config,
    connect_to_bigquery,
    get_schema_column_info_big_query,
    get_tables_by_schema_big_query,
    run_query_big_query,
)


def _is_postgres(database_type):
    return database_type and database_type.lower() in ("postgresql", "postgres")


def get_db_credentials(database_type, connection_string):
    if _is_postgres(database_type):
        return format_postgres(connection_string)
    elif database_type and database_type.lower() == "bigquery":
        return format_bigquery_config(connection_string)
    return {}


def connect_to_db(database_type, config, using_connection_string):
    if _is_postgres(database_type):
        return connect_to_postgres(config, using_connection_string)
    elif database_type and database_type.lower() == "bigquery":
        return connect_to_bigquery(config, using_connection_string)
    return None


def create_async_connection(database_type, config, using_connection_string):
    if _is_postgres(database_type):
        return connect_to_postgres_async(config, using_connection_string)
    return connect_to_db(database_type, config, using_connection_string)


async def connect_to_db_async(database_type, config, using_connection_string):
    connection = create_async_connection(
        database_type, config, using_connection_string
    )
    if _is_postgres(database_type):
        await connection.open(wait=False)
        return connection
    return connection


def run_query_by_db(database_type, query, connection):
    if _is_postgres(database_type):
        return run_query_postgres(query, connection)
    elif database_type and database_type.lower() == "bigquery":
        return run_query_big_query(query, connection)
    return None


async def run_query_by_db_async(database_type, query, connection):
    if _is_postgres(database_type):
        return await run_query_postgres_async(query, connection)
    return run_query_by_db(database_type, query, connection)


def disconnect_from_db(database_type, connection):
    if _is_postgres(database_type):
        return disconnect_from_postgres(connection)
    return None


async def disconnect_from_db_async(database_type, connection):
    if _is_postgres(database_type):
        return await disconnect_from_postgres_async(connection)
    return disconnect_from_db(database_type, connection)


def get_schema_tables_by_db(database_type, connection, schema_name):
    if _is_postgres(database_type):
        return get_tables_by_schema_postgres(connection, schema_name)
    elif database_type and database_type.lower() == "bigquery":
        return get_tables_by_schema_big_query(connection, schema_name)
    return None


def get_schema_column_info_by_db(database_type, connection, schema_name, table_names):
    if _is_postgres(database_type):
        return get_schema_column_info_postgres(connection, schema_name, table_names)
    elif database_type and database_type.lower() == "bigquery":
        return get_schema_column_info_big_query(connection, schema_name, table_names)
    return None



from __future__ import annotations

import threading
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Callable, Optional


class ParallelPostQuill:
    """
    Provides a thread-pooled, drop-in replacement for Quill.post_quill.

    Example
    -------
        parallel_post = ParallelPostQuill(
            quill_factory=lambda: Quill(...),
            max_workers=4,
        )

        # Synchronous (behaves like the original post_quill)
        response = parallel_post("report", payload)

        # Async-style execution
        future = parallel_post("report", payload, wait=False)
        response = future.result()
    """

    def __init__(
        self,
        quill_factory: Callable[[], "Quill"],
        max_workers: int = 5,
        executor: Optional[ThreadPoolExecutor] = None,
    ):
        if executor is None:
            self._executor = ThreadPoolExecutor(
                max_workers=max_workers,
                thread_name_prefix="post-quill",
            )
            self._owns_executor = True
        else:
            self._executor = executor
            self._owns_executor = False

        if not callable(quill_factory):
            raise ValueError("quill_factory must be callable and create Quill instances.")

        self._quill_factory = quill_factory
        self._thread_local = threading.local()

    def __call__(self, path: str, payload: dict, wait: bool = True):
        future: Future = self._executor.submit(self._invoke, path, payload)
        if wait:
            return future.result()
        return future

    def shutdown(self, wait: bool = True):
        if self._owns_executor:
            self._executor.shutdown(wait=wait)

    def _get_quill(self):
        quill = getattr(self._thread_local, "quill", None)
        if quill is None:
            quill = self._quill_factory()
            self._thread_local.quill = quill
        return quill

    def _invoke(self, path: str, payload: dict):
        quill = self._get_quill()
        return quill.post_quill(path, payload)




# __init__.py

from .core import Quill
from .error import PgQueryError
from .utils import (
    Filter,
    FilterType,
    FieldType,
    StringOperator,
    DateOperator,
    NumberOperator,
    NullOperator,
    ParallelPostQuill,
)



from typing import Union, List, Dict

# Type aliases for clarity
TenantId = Union[str, int]
TenantInfo = Dict[str, Union[str, List[TenantId]]]
Tenants = Union[List[TenantId], List[TenantInfo]]

def extract_tenant_ids(tenants: Tenants) -> List[TenantId]:
    """
    Extract tenant IDs from the tenants parameter, which can be either a list of IDs
    or a list of tenant info dictionaries.
    
    Args:
        tenants: Either a list of tenant IDs (strings/integers) or a list of tenant info dictionaries
        
    Returns:
        List of tenant IDs
        
    Raises:
        ValueError: If the tenants parameter format is invalid
    """
    if not tenants:
        raise ValueError("Invalid format for tenants: empty list")

    first_tenant = tenants[0]
    
    if isinstance(first_tenant, (str, int)):
        return tenants  # type: ignore
    elif isinstance(first_tenant, dict) and "tenantIds" in first_tenant:
        # TODO: support multiple tenants in future
        return first_tenant["tenantIds"]
    else:
        raise ValueError("Invalid format for tenants")

def extract_tenant_field(tenants: Tenants, dashboard_owner: str) -> str:
    """
    Extract tenant field from the tenants parameter, falling back to dashboard_owner
    if tenants is a simple list of IDs.
    
    Args:
        tenants: Either a list of tenant IDs (strings/integers) or a list of tenant info dictionaries
        dashboard_owner: The default tenant field to use if tenants is a simple list
        
    Returns:
        The tenant field string
        
    Raises:
        ValueError: If the tenants parameter format is invalid
    """
    if not tenants:
        raise ValueError("Invalid format for tenants: empty list")

    first_tenant = tenants[0]
    
    if isinstance(first_tenant, (str, int)):
        return dashboard_owner
    elif isinstance(first_tenant, dict) and "tenantField" in first_tenant:
        return first_tenant["tenantField"]
    else:
        raise ValueError("Invalid format for tenants") 


import os
import codecs
import re
import hashlib
import asyncio
import inspect
from concurrent.futures import ThreadPoolExecutor
from dotenv import load_dotenv

import requests
import redis
from .utils import Filter, convert_custom_filter
import json
from enum import Enum


from quillsql.db.cached_connection import CachedConnection
from quillsql.db.db_helper import (
    get_db_credentials,
    get_schema_column_info_by_db,
    get_schema_tables_by_db,
)
from quillsql.utils.schema_conversion import convert_type_to_postgres
from quillsql.utils.run_query_processes import (
    array_to_map,
    remove_fields,
)
from quillsql.utils.tenants import extract_tenant_ids
from quillsql.utils.pivot_template import (
    parse_distinct_values,
    hydrate_pivot_template,
)

load_dotenv()

ENV = os.getenv("PYTHON_ENV")
DEV_HOST = "http://localhost:8080"
PROD_HOST = "https://api.quill.co"
HOST = DEV_HOST if ENV == "development" else PROD_HOST

SINGLE_TENANT = "QUILL_SINGLE_TENANT"
ALL_TENANTS = "QUILL_ALL_TENANTS"
FLAG_TASKS = {'dashboard', 'report', 'item', 'report-info', 'filter-options'}

# Quill - Fullstack API Platform for Dashboards and Reporting.
class Quill:
    def __init__(
        self,
        private_key,
        database_type,
        database_connection_string=None,
        database_config=None,
        metadataServerURL=None,
        cache=None,
    ):
        if private_key is None:
            raise ValueError("Private key is required")
        if database_type is None:
            raise ValueError("Database type is required")
        if database_connection_string is None and database_config is None:
            raise ValueError("You must provide either DatabaseConnectionString or DatabaseConfig")

        # Handles both dsn-style connection strings (eg. "dbname=test password=secret" )
        # as well as url-style connection strings (eg. "postgres://foo@db.com")
        self.baseUrl = metadataServerURL if metadataServerURL is not None else HOST
        if database_connection_string is not None:
            self.target_connection = CachedConnection(
                database_type,
                get_db_credentials(database_type, database_connection_string),
                cache,
                True,
            )
        else:
            self.target_connection = CachedConnection(
                database_type, database_config, cache, False
            )
        self.private_key = private_key

    def close(self):
        connection = getattr(self, "target_connection", None)
        if connection and hasattr(connection, "close"):
            connection.close()

    async def aclose(self):
        connection = getattr(self, "target_connection", None)
        if connection and hasattr(connection, "close_async"):
            await connection.close_async()
            return
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aclose()
        return False

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def get_cache(self, cache_config):
        cache_type = cache_config and cache_config.get("cache_type")
        if cache_type and cache_type == "redis" or cache_type == "rediss":
            return redis.Redis(
                host=cache_config.get("host", "localhost"),
                port=cache_config.get("port", 6379),
                username=cache_config.get("username", "default"),
                password=cache_config.get("password"),
            )
        return None

    def _get_tenant_mapped_flags_cache_key(self, client_id, tenants, report_id):
        if not client_id or not isinstance(tenants, list) or not tenants:
            return None
        if not isinstance(tenants[0], dict):
            return None

        normalized_tenants = sorted(
            [
                {
                    "tenantField": tenant.get("tenantField"),
                    "tenantIds": sorted(
                        list(tenant.get("tenantIds") or []), key=lambda value: str(value)
                    ),
                }
                for tenant in tenants
                if isinstance(tenant, dict)
            ],
            key=lambda tenant: tenant.get("tenantField") or "",
        )
        return (
            f"tenant-mapped-flags:{client_id}:{report_id or ''}:"
            f"{json.dumps(normalized_tenants, separators=(',', ':'), ensure_ascii=False)}"
        )

    def _get_tenant_mapped_flags_from_cache(self, cache_key, overwrite_cache):
        cache = getattr(self.target_connection, "cache", None)
        if overwrite_cache or not cache_key or not cache:
            return None
        try:
            cached_response = cache.get(cache_key)
            if not cached_response:
                return None
            parsed_response = json.loads(cached_response)
            is_valid_cache = (
                isinstance(parsed_response, list)
                and all(
                    isinstance(item, dict)
                    and isinstance(item.get("tenantField"), str)
                    and isinstance(item.get("flags"), list)
                    for item in parsed_response
                )
            )
            if not is_valid_cache:
                try:
                    cache.delete(cache_key)
                except Exception:
                    pass
                return None
            return parsed_response
        except Exception:
            try:
                if cache and cache_key:
                    cache.delete(cache_key)
            except Exception:
                pass
            return None

    def _set_tenant_mapped_flags_cache(self, cache_key, tenant_flags):
        cache = getattr(self.target_connection, "cache", None)
        if not cache_key or not cache:
            return
        ttl = getattr(self.target_connection, "ttl", None)
        try:
            if ttl:
                cache.set(cache_key, json.dumps(tenant_flags), "EX", ttl)
            else:
                cache.set(cache_key, json.dumps(tenant_flags))
        except Exception:
            pass

    def _post_quill_sync(self, path, payload, use_cache=True):
        if use_cache:
            result = self.post_quill(path, payload)
        else:
            result = self.post_quill(path, payload, use_cache)
        if inspect.isawaitable(result):
            try:
                asyncio.get_running_loop()
            except RuntimeError:
                return asyncio.run(result)
            with ThreadPoolExecutor(max_workers=1) as executor:
                return executor.submit(lambda: asyncio.run(result)).result()
        return result

    async def query(
        self,
        tenants,
        metadata,
        flags=None,
        filters: list[Filter] = None,
        admin_enabled: bool = None,
    ):
        if getattr(self.target_connection, "database_type", "").lower() in (
            "postgres",
            "postgresql",
        ):
            return await self._query_async(
                tenants, metadata, flags, filters, admin_enabled
            )
        return await asyncio.to_thread(
            self._query_sync,
            tenants,
            metadata,
            flags,
            filters,
            admin_enabled,
        )

    def query_sync(
        self,
        tenants,
        metadata,
        flags=None,
        filters: list[Filter] = None,
        admin_enabled: bool = None,
    ):
        return self._query_sync(tenants, metadata, flags, filters, admin_enabled)

    async def _query_async(
        self,
        tenants,
        metadata,
        flags=None,
        filters: list[Filter] = None,
        admin_enabled: bool = None,
    ):
        if not tenants:
            raise ValueError("You may not pass an empty tenants array.")

        responseMetadata = {}
        if not metadata:
            return {"error": "Missing metadata.", "status": "error", "data": {}}

        task = metadata.get("task")
        if not task:
            return {"error": "Missing task.", "status": "error", "data": {}}
        overwrite_cache = bool(metadata.get("overwriteCache", False))

        try:
            self.target_connection.tenant_ids = extract_tenant_ids(tenants)

            if task == "pivot-template":
                pivot_payload = {
                    **metadata,
                    "tenants": tenants,
                    "flags": flags,
                }
                if admin_enabled is not None:
                    pivot_payload["adminEnabled"] = admin_enabled
                if filters is not None:
                    pivot_payload["sdkFilters"] = [
                        convert_custom_filter(f) for f in filters
                    ]
                pivot_template_response = await self.post_quill(
                    "pivot-template",
                    pivot_payload,
                )

                if pivot_template_response.get("error"):
                    return {
                        "status": "error",
                        "error": pivot_template_response.get("error"),
                        "data": pivot_template_response.get("metadata") or {},
                    }

                template = pivot_template_response.get("metadata", {}).get("template")
                config = pivot_template_response.get("metadata", {}).get("config")
                distinct_values_query = pivot_template_response.get(
                    "metadata", {}
                ).get("distinctValuesQuery")
                row_count_query = pivot_template_response.get("metadata", {}).get(
                    "rowCountQuery"
                )

                distinct_values = []
                if distinct_values_query:
                    distinct_value_results = await self.run_queries_async(
                        [distinct_values_query],
                        self.target_connection.database_type,
                        metadata.get("databaseType"),
                        metadata,
                        None,
                        overwrite_cache,
                    )

                    distinct_value_query_results = (
                        distinct_value_results or {}
                    ).get("queryResults") or []
                    if distinct_value_query_results:
                        distinct_values = parse_distinct_values(
                            distinct_value_query_results[0], config.get("databaseType")
                        )

                try:
                    final_query = hydrate_pivot_template(template, distinct_values, config)
                except Exception as err:
                    return {
                        "status": "error",
                        "error": f"Failed to hydrate pivot template: {str(err)}",
                        "data": {},
                    }

                queries_to_run = [final_query]
                if row_count_query:
                    hydrated_row_count_query = hydrate_pivot_template(
                        row_count_query,
                        distinct_values,
                        config,
                    )
                    queries_to_run.append(hydrated_row_count_query)

                final_results = await self.run_queries_async(
                    queries_to_run,
                    self.target_connection.database_type,
                    metadata.get("databaseType"),
                    metadata,
                    pivot_template_response.get("metadata", {}).get("runQueryConfig"),
                    overwrite_cache,
                )

                responseMetadata = pivot_template_response.get("metadata") or {}
                if (
                    final_results.get("queryResults")
                    and len(final_results["queryResults"]) >= 1
                ):
                    query_results = final_results["queryResults"][0]
                    if query_results.get("rows"):
                        responseMetadata["rows"] = query_results["rows"]
                    if query_results.get("fields"):
                        responseMetadata["fields"] = query_results["fields"]

                if "template" in responseMetadata:
                    del responseMetadata["template"]
                if "distinctValuesQuery" in responseMetadata:
                    del responseMetadata["distinctValuesQuery"]
                if "rowCountQuery" in responseMetadata:
                    del responseMetadata["rowCountQuery"]

                return {
                    "data": responseMetadata,
                    "queries": final_results,
                    "status": "success",
                }

            tenant_flags = None
            if (
                task in FLAG_TASKS
                and tenants[0] != ALL_TENANTS
                and tenants[0] != SINGLE_TENANT
            ):
                tenant_mapped_flags_cache_key = self._get_tenant_mapped_flags_cache_key(
                    metadata.get("clientId"),
                    tenants,
                    metadata.get("reportId") or "",
                )
                cached_tenant_flags = self._get_tenant_mapped_flags_from_cache(
                    tenant_mapped_flags_cache_key,
                    overwrite_cache,
                )
                if cached_tenant_flags:
                    tenant_flags = cached_tenant_flags
                else:
                    tenant_flags_payload = {
                        "reportId": metadata.get("reportId")
                        or metadata.get("dashboardItemId"),
                        "clientId": metadata.get("clientId"),
                        "dashboardName": metadata.get("name"),
                        "tenants": tenants,
                        "flags": flags,
                    }
                    if admin_enabled is not None:
                        tenant_flags_payload["adminEnabled"] = admin_enabled
                    response = await self.post_quill(
                        "tenant-mapped-flags", tenant_flags_payload
                    )

                    if response.get("error"):
                        return {
                            "status": "error",
                            "error": response.get("error"),
                            "data": response.get("metadata") or {},
                        }

                    flag_query_results = await self.run_queries_async(
                        response.get("queries"),
                        self.target_connection.database_type,
                        overwrite_cache=overwrite_cache,
                    )

                    tenant_flags = []
                    query_order = (response.get("metadata") or {}).get("queryOrder") or []
                    query_results = (flag_query_results or {}).get("queryResults") or []
                    for index, tenant_field in enumerate(query_order):
                        query_result = (
                            query_results[index] if index < len(query_results) else {}
                        )
                        rows = (
                            query_result.get("rows")
                            if isinstance(query_result, dict)
                            else []
                        )
                        ordered_flags = []
                        seen = set()
                        for row in rows or []:
                            if not isinstance(row, dict):
                                continue
                            flag = row.get("quill_flag")
                            if flag is None or flag in seen:
                                continue
                            seen.add(flag)
                            ordered_flags.append(flag)
                        tenant_flags.append(
                            {
                                "tenantField": tenant_field,
                                "flags": ordered_flags,
                            }
                        )
                    self._set_tenant_mapped_flags_cache(
                        tenant_mapped_flags_cache_key,
                        tenant_flags,
                    )

            elif tenants[0] == SINGLE_TENANT and flags:
                if flags and isinstance(flags[0], dict):
                    tenant_flags = [{"tenantField": SINGLE_TENANT, "flags": flags}]
                else:
                    tenant_flags = flags

            if metadata.get("preQueries"):
                pre_query_results = await self.run_queries_async(
                    metadata.get("preQueries"),
                    self.target_connection.database_type,
                    metadata.get("databaseType"),
                    metadata,
                    metadata.get("runQueryConfig"),
                    overwrite_cache,
                )
            else:
                pre_query_results = {}

            if metadata.get("runQueryConfig") and metadata.get("runQueryConfig").get(
                "overridePost"
            ):
                return {"data": pre_query_results, "status": "success"}
            view_query = None
            if metadata.get("preQueries"):
                view_query = metadata.get("preQueries")[0]
            pre_query_columns = (
                pre_query_results.get("columns")
                if metadata.get("runQueryConfig")
                and metadata.get("runQueryConfig").get("getColumns")
                else None
            )
            payload = {
                **metadata,
                "tenants": tenants,
                "flags": tenant_flags,
                "viewQuery": view_query,
            }
            if pre_query_columns is not None:
                payload["preQueryResultsColumns"] = pre_query_columns
            if admin_enabled is not None:
                payload["adminEnabled"] = admin_enabled
            if filters is not None:
                payload["sdkFilters"] = [convert_custom_filter(f) for f in filters]
            quill_results = await self.post_quill(metadata.get("task"), payload)
            if quill_results.get("error"):
                responseMetadata = quill_results.get("metadata")
                response = {
                    "error": quill_results.get("error"),
                    "status": "error",
                    "data": {},
                }
                if responseMetadata:
                    response["data"] = responseMetadata
                return response

            if not quill_results.get("metadata"):
                quill_results["metadata"] = {}
            metadata = quill_results.get("metadata")
            responseMetadata = metadata
            results = await self.run_queries_async(
                quill_results.get("queries"),
                self.target_connection.database_type,
                metadata.get("databaseType"),
                metadata,
                metadata.get("runQueryConfig"),
                overwrite_cache,
            )

            should_wrap_results = isinstance(results, list) or not results
            if should_wrap_results:
                normalized_results = {
                    "queryResults": results if isinstance(results, list) else []
                }
            else:
                normalized_results = results

            if (
                should_wrap_results
                and not normalized_results.get("queryResults")
                and quill_results.get("queries")
            ):
                normalized_results["queryResults"] = (
                    normalized_results.get("queryResults") or []
                )

            if (
                normalized_results.get("mapped_array")
                and metadata.get("runQueryConfig", {}).get("arrayToMap")
            ):
                array_to_map = metadata["runQueryConfig"]["arrayToMap"]
                target_collection = responseMetadata.get(array_to_map["arrayName"])
                if isinstance(target_collection, list):
                    for index, array in enumerate(normalized_results["mapped_array"]):
                        if index >= len(target_collection):
                            continue
                        target_entry = target_collection[index]
                        if not isinstance(target_entry, dict):
                            target_entry = {}
                            target_collection[index] = target_entry
                        target_entry[array_to_map["field"]] = (
                            array if array is not None else []
                        )
                del normalized_results["mapped_array"]

            query_results_list = normalized_results.get("queryResults") or []
            if len(query_results_list) == 1 and isinstance(query_results_list[0], dict):
                query_result = query_results_list[0]
                quill_results["metadata"]["rows"] = query_result.get("rows")
                quill_results["metadata"]["fields"] = query_result.get("fields")
            return {
                "data": quill_results.get("metadata"),
                "queries": normalized_results,
                "status": "success",
            }

        except Exception as err:
            if task == "update-view":
                broken_view_payload = {
                    "table": metadata.get("name"),
                    "clientId": metadata.get("clientId"),
                    "error": str(err),
                }
                if admin_enabled is not None:
                    broken_view_payload["adminEnabled"] = admin_enabled
                await self.post_quill("set-broken-view", broken_view_payload)
            return {
                "error": str(err).splitlines()[0],
                "status": "error",
                "data": responseMetadata,
            }

    def _query_sync(
        self,
        tenants,
        metadata,
        flags=None,
        filters: list[Filter] = None,
        admin_enabled: bool = None,
    ):
        if not tenants:
            raise ValueError("You may not pass an empty tenants array.")

        responseMetadata = {}
        if not metadata:
            return {"error": "Missing metadata.", "status": "error", "data": {}}

        task = metadata.get("task")
        if not task:
            return {"error": "Missing task.", "status": "error", "data": {}}
        overwrite_cache = bool(metadata.get("overwriteCache", False))

        try:
            # Set tenant IDs in the connection
            self.target_connection.tenant_ids = extract_tenant_ids(tenants)

            # Handle pivot-template task
            if task == "pivot-template":
                pivot_payload = {
                    **metadata,
                    "tenants": tenants,
                    "flags": flags,
                }
                if admin_enabled is not None:
                    pivot_payload["adminEnabled"] = admin_enabled
                if filters is not None:
                    pivot_payload["sdkFilters"] = [
                        convert_custom_filter(f) for f in filters
                    ]
                pivot_template_response = self._post_quill_sync(
                    "pivot-template",
                    pivot_payload,
                )

                if pivot_template_response.get("error"):
                    return {
                        "status": "error",
                        "error": pivot_template_response.get("error"),
                        "data": pivot_template_response.get("metadata") or {},
                    }

                template = pivot_template_response.get("metadata", {}).get("template")
                config = pivot_template_response.get("metadata", {}).get("config")
                distinct_values_query = pivot_template_response.get("metadata", {}).get("distinctValuesQuery")
                row_count_query = pivot_template_response.get("metadata", {}).get("rowCountQuery")

                distinct_values = []
                if distinct_values_query:
                    distinct_value_results = self.run_queries(
                        [distinct_values_query],
                        self.target_connection.database_type,
                        metadata.get("databaseType"),
                        metadata,
                        None,
                        overwrite_cache,
                    )

                    distinct_value_query_results = (
                        distinct_value_results or {}
                    ).get("queryResults") or []
                    if distinct_value_query_results:
                        distinct_values = parse_distinct_values(
                            distinct_value_query_results[0],
                            config.get("databaseType")
                        )

                try:
                    final_query = hydrate_pivot_template(template, distinct_values, config)
                except Exception as err:
                    return {
                        "status": "error",
                        "error": f"Failed to hydrate pivot template: {str(err)}",
                        "data": {},
                    }

                queries_to_run = [final_query]
                if row_count_query:
                    hydrated_row_count_query = hydrate_pivot_template(
                        row_count_query,
                        distinct_values,
                        config
                    )
                    queries_to_run.append(hydrated_row_count_query)

                final_results = self.run_queries(
                    queries_to_run,
                    self.target_connection.database_type,
                    metadata.get("databaseType"),
                    metadata,
                    pivot_template_response.get("metadata", {}).get("runQueryConfig"),
                    overwrite_cache,
                )

                responseMetadata = pivot_template_response.get("metadata") or {}
                if final_results.get("queryResults") and len(final_results["queryResults"]) >= 1:
                    query_results = final_results["queryResults"][0]
                    if query_results.get("rows"):
                        responseMetadata["rows"] = query_results["rows"]
                    if query_results.get("fields"):
                        responseMetadata["fields"] = query_results["fields"]

                if "template" in responseMetadata:
                    del responseMetadata["template"]
                if "distinctValuesQuery" in responseMetadata:
                    del responseMetadata["distinctValuesQuery"]
                if "rowCountQuery" in responseMetadata:
                    del responseMetadata["rowCountQuery"]

                return {
                    "data": responseMetadata,
                    "queries": final_results,
                    "status": "success",
                }

            # Handle tenant flags synthesis
            tenant_flags = None
            if (task in FLAG_TASKS and 
                tenants[0] != ALL_TENANTS and 
                tenants[0] != SINGLE_TENANT
            ):
                tenant_mapped_flags_cache_key = self._get_tenant_mapped_flags_cache_key(
                    metadata.get("clientId"),
                    tenants,
                    metadata.get("reportId") or "",
                )
                cached_tenant_flags = self._get_tenant_mapped_flags_from_cache(
                    tenant_mapped_flags_cache_key,
                    overwrite_cache,
                )
                if cached_tenant_flags:
                    tenant_flags = cached_tenant_flags
                else:
                    tenant_flags_payload = {
                        'reportId': metadata.get('reportId') or metadata.get('dashboardItemId'),
                        'clientId': metadata.get('clientId'),
                        'dashboardName': metadata.get('name'),
                        'tenants': tenants,
                        'flags': flags,
                    }
                    if admin_enabled is not None:
                        tenant_flags_payload['adminEnabled'] = admin_enabled
                    response = self._post_quill_sync('tenant-mapped-flags', tenant_flags_payload)
                    
                    if response.get('error'):
                        return {
                            'status': 'error',
                            'error': response.get('error'),
                            'data': response.get('metadata') or {},
                        }
                    
                    flag_query_results = self.run_queries(
                        response.get('queries'),
                        self.target_connection.database_type,
                        overwrite_cache=overwrite_cache,
                    )
                    
                    tenant_flags = []
                    query_order = (response.get('metadata') or {}).get('queryOrder') or []
                    query_results = (flag_query_results or {}).get('queryResults') or []
                    for index, tenant_field in enumerate(query_order):
                        query_result = (
                            query_results[index]
                            if index < len(query_results)
                            else {}
                        )
                        rows = (
                            query_result.get('rows')
                            if isinstance(query_result, dict)
                            else []
                        )
                        ordered_flags = []
                        seen = set()
                        for row in rows or []:
                            if not isinstance(row, dict):
                                continue
                            flag = row.get('quill_flag')
                            if flag is None or flag in seen:
                                continue
                            seen.add(flag)
                            ordered_flags.append(flag)
                        tenant_flags.append({
                            'tenantField': tenant_field,
                            'flags': ordered_flags,
                        })
                    self._set_tenant_mapped_flags_cache(
                        tenant_mapped_flags_cache_key,
                        tenant_flags,
                    )

            elif tenants[0] == SINGLE_TENANT and flags:
                if flags and isinstance(flags[0], dict):
                    tenant_flags = [{'tenantField': SINGLE_TENANT, 'flags': flags}]
                else:
                    tenant_flags = flags

            if metadata.get("preQueries"):
                pre_query_results = self.run_queries(
                    metadata.get("preQueries"),
                    self.target_connection.database_type,
                    metadata.get("databaseType"),
                    metadata,
                    metadata.get("runQueryConfig"),
                    overwrite_cache,
                )
            else:
                pre_query_results = {}

            if metadata.get("runQueryConfig") and metadata.get("runQueryConfig").get(
                "overridePost"
            ):
                return {"data": pre_query_results, "status": "success"}
            view_query = None
            if metadata.get("preQueries"):
                view_query = metadata.get("preQueries")[0]
            pre_query_columns = (
                pre_query_results.get("columns")
                if metadata.get("runQueryConfig")
                and metadata.get("runQueryConfig").get("getColumns")
                else None
            )
            payload = {
                **metadata,
                "tenants": tenants,
                "flags": tenant_flags,
                "viewQuery": view_query,
            }
            if pre_query_columns is not None:
                payload["preQueryResultsColumns"] = pre_query_columns
            if admin_enabled is not None:
                payload["adminEnabled"] = admin_enabled
            if filters is not None:
                payload["sdkFilters"] = [convert_custom_filter(f) for f in filters]
            quill_results = self._post_quill_sync(metadata.get("task"), payload)
            if quill_results.get("error"):
                responseMetadata = quill_results.get("metadata")
                response = {
                    "error": quill_results.get("error"),
                    "status": "error",
                    "data": {},
                }
                if responseMetadata:
                    response["data"] = responseMetadata
                return response

            # If there is no metadata in the quill results, create one
            if not quill_results.get("metadata"):
                quill_results["metadata"] = {}
            metadata = quill_results.get("metadata")
            responseMetadata = metadata
            results = self.run_queries(
                quill_results.get("queries"),
                self.target_connection.database_type,
                metadata.get("databaseType"),
                metadata,
                metadata.get("runQueryConfig"),
                overwrite_cache,
            )

            should_wrap_results = isinstance(results, list) or not results
            if should_wrap_results:
                normalized_results = {
                    "queryResults": results if isinstance(results, list) else []
                }
            else:
                normalized_results = results

            if (
                should_wrap_results
                and not normalized_results.get("queryResults")
                and quill_results.get("queries")
            ):
                normalized_results["queryResults"] = (
                    normalized_results.get("queryResults") or []
                )

            if (
                normalized_results.get("mapped_array")
                and metadata.get("runQueryConfig", {}).get("arrayToMap")
            ):
                array_to_map = metadata["runQueryConfig"]["arrayToMap"]
                target_collection = responseMetadata.get(array_to_map["arrayName"])
                if isinstance(target_collection, list):
                    for index, array in enumerate(normalized_results["mapped_array"]):
                        if index >= len(target_collection):
                            continue
                        target_entry = target_collection[index]
                        if not isinstance(target_entry, dict):
                            target_entry = {}
                            target_collection[index] = target_entry
                        target_entry[array_to_map["field"]] = array if array is not None else []
                del normalized_results["mapped_array"]

            query_results_list = normalized_results.get("queryResults") or []
            if len(query_results_list) == 1 and isinstance(query_results_list[0], dict):
                query_result = query_results_list[0]
                quill_results["metadata"]["rows"] = query_result.get("rows")
                quill_results["metadata"]["fields"] = query_result.get("fields")
            return {
                "data": quill_results.get("metadata"),
                "queries": normalized_results,
                "status": "success",
            }

        except Exception as err:
            if task == "update-view":
                broken_view_payload = {
                    "table": metadata.get("name"),
                    "clientId": metadata.get("clientId"),
                    "error": str(err),
                }
                if admin_enabled is not None:
                    broken_view_payload["adminEnabled"] = admin_enabled
                self._post_quill_sync("set-broken-view", broken_view_payload)
            return {
                "error": str(err).splitlines()[0],
                "status": "error",
                "data": responseMetadata,
            }
        
    async def stream(
        self,
        tenants,
        metadata,
        flags=None,
        filters=None,
        admin_enabled=None,
    ):
        if not tenants:
            raise ValueError("You may not pass an empty tenants array.")

        if not metadata:
            yield {"type": "error", "errorText": "Missing metadata."}
            return

        task = metadata.get("task")
        if not task:
            yield {"type": "error", "errorText": "Missing task."}
            return

        try:
            # Set tenant IDs in the connection
            self.target_connection.tenant_ids = extract_tenant_ids(tenants)
            if task in ("chat", "agent"):
                for event in self._agentic_chat_loop(
                    tenants,
                    metadata,
                    flags,
                    filters,
                    admin_enabled,
                ):
                    yield event
                return

            for event in self._stream_sse(
                task,
                tenants,
                metadata,
                flags,
                filters,
                admin_enabled,
            ):
                yield event
            return
        except Exception as err:
            yield {
                "type": "error",
                "errorText": str(err).splitlines()[0],
            }
            return

    def _normalize_tenant_flags(self, tenants, flags):
        tenant_flags = None
        if tenants and tenants[0] == SINGLE_TENANT and flags:
            if flags and isinstance(flags[0], dict):
                tenant_flags = [{"tenantField": SINGLE_TENANT, "flags": flags}]
            else:
                tenant_flags = flags
        return tenant_flags

    def _agentic_chat_loop(self, tenants, metadata, flags, filters, admin_enabled):
        messages = list(metadata.get("messages") or [])
        max_iterations = 10

        for _ in range(max_iterations):
            payload = {
                **metadata,
                "messages": messages,
            }

            has_tool_calls = False
            assistant_text = ""
            tool_results = []
            tool_calls = []

            for event in self._stream_sse(
                "agent",
                tenants,
                payload,
                flags,
                filters,
                admin_enabled,
            ):
                yield event

                if event.get("type") == "text-delta":
                    assistant_text += event.get("delta", "")

                if event.get("type") == "tool-input-available":
                    tool_name = event.get("toolName")
                    tool_call_id = event.get("toolCallId")
                    tool_input = event.get("input") or {}

                    if tool_call_id is None:
                        yield {
                            "type": "error",
                            "errorText": "Missing toolCallId for tool-input-available event.",
                        }
                        continue

                    has_tool_calls = True
                    yield {
                        "type": "tool-executing",
                        "toolCallId": tool_call_id,
                        "toolName": tool_name,
                    }

                    result = self._execute_tool_locally(tool_name, tool_input)
                    tool_results.append(
                        {
                            "toolCallId": tool_call_id,
                            "toolName": tool_name,
                            "input": tool_input,
                            "result": result,
                        }
                    )
                    tool_calls.append(
                        {
                            "id": tool_call_id,
                            "type": "function",
                            "function": {
                                "name": tool_name,
                                "arguments": json.dumps(tool_input or {}),
                            },
                        }
                    )

                    yield {
                        "type": "tool-result",
                        "toolCallId": tool_call_id,
                        "result": result,
                    }

                if event.get("type") in ("finish", "error"):
                    break

            if not has_tool_calls:
                break

            def _build_tool_result_for_history(tool_result):
                result = tool_result.get("result") or {}
                has_rows = isinstance(result.get("rows"), list)
                has_fields = isinstance(result.get("fields"), list)
                is_query_result = has_rows or has_fields or result.get("dbMismatched")
                if not is_query_result:
                    return result
                tool_input = tool_result.get("input") or {}
                error = result.get("error") or tool_input.get("error")
                status = "error" if error or result.get("dbMismatched") else "success"
                payload = {"status": status}
                if tool_input.get("sql"):
                    payload["sql"] = tool_input.get("sql")
                if error:
                    payload["error"] = error
                if result.get("dbMismatched"):
                    payload["meta"] = {
                        "dbMismatched": True,
                        "backendDatabaseType": result.get("backendDatabaseType"),
                    }
                elif not error:
                    payload["meta"] = {
                        "rowsFetchedSuccessfully": True,
                        "rowCount": len(result.get("rows") or []),
                    }
                return payload

            messages.append(
                {
                    "role": "assistant",
                    "content": assistant_text or None,
                    "tool_calls": tool_calls,
                }
            )
            for tool_result in tool_results:
                messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_result["toolCallId"],
                        "content": json.dumps(_build_tool_result_for_history(tool_result)),
                    }
                )

        yield {"type": "done"}

    def _execute_tool_locally(self, tool_name, tool_input):
        if tool_name == "generateReport":
            if tool_input.get("error"):
                return {"error": tool_input.get("error")}
            sql = tool_input.get("sql")
            if not sql:
                return {"error": "No SQL provided"}
            results = self.run_queries(
                [sql],
                self.target_connection.database_type,
            )
            if results.get("dbMismatched"):
                return results
            query_results = results.get("queryResults") or []
            if query_results and isinstance(query_results[0], dict):
                if query_results[0].get("error"):
                    return query_results[0]
                return {
                    "rows": query_results[0].get("rows", []),
                    "fields": query_results[0].get("fields", []),
                }
            return {"rows": [], "fields": []}

        if tool_name == "createChart":
            return {"chartConfig": tool_input}

        return {"error": f"Unknown tool: {tool_name}"}

    def _stream_sse(self, endpoint, tenants, payload, flags, filters, admin_enabled):
        tenant_flags = self._normalize_tenant_flags(tenants, flags)
        request_payload = {
            **payload,
            "tenants": tenants,
            "flags": tenant_flags,
        }
        if filters:
            request_payload["sdkFilters"] = [convert_custom_filter(f) for f in filters]
        if admin_enabled is not None:
            request_payload["adminEnabled"] = admin_enabled

        # Custom JSON Encoder to handle Enums
        class EnumEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, Enum):
                    return obj.value  # Convert enum to its value (string in this case)
                return super().default(obj)

        url = f"{self.baseUrl}/sdk/{endpoint}"
        headers = {
            "Authorization": f"Bearer {self.private_key}",
            "Content-Type": "application/json",
            "Accept": "text/event-stream",
        }
        encoded = json.dumps(request_payload, cls=EnumEncoder)

        resp = requests.post(url, data=encoded, headers=headers, stream=True)
        decoder = codecs.getincrementaldecoder("utf-8")()
        buf = ""
        for chunk in resp.iter_content(chunk_size=4096):
            buf += decoder.decode(chunk)
            while "\n\n" in buf:
                raw_event, buf = buf.split("\n\n", 1)
                data_lines = []
                for line in raw_event.splitlines():
                    if line.startswith("data:"):
                        data_lines.append(line[len("data:"):].strip())
                if not data_lines:
                    continue
                payload = "\n".join(data_lines)
                if payload == "[DONE]":
                    return
                try:
                    parsed = json.loads(payload)
                    if isinstance(parsed, str):
                        yield {"type": "text-delta", "id": "0", "delta": parsed}
                    else:
                        yield parsed
                except json.JSONDecodeError:
                    continue

        # flush any partial code points at the end
        buf += decoder.decode(b"", final=True)
        yield buf
        return

    def apply_limit(self, query, limit):
        # Simple logic: if query already has a limit, don't add another
        if getattr(self.target_connection, 'database_type', '').lower() == 'mssql':
            import re
            if re.search(r'SELECT TOP \\d+', query, re.IGNORECASE):
                return query
            return re.sub(r'select', f'SELECT TOP {limit}', query, flags=re.IGNORECASE)
        else:
            if 'limit ' in query.lower():
                return query
            return f"{query.rstrip(';')} limit {limit}"

    async def run_queries_async(
        self,
        queries,
        pkDatabaseType,
        databaseType=None,
        metadata=None,
        runQueryConfig=None,
        overwrite_cache=False,
    ):
        if getattr(self.target_connection, "database_type", "").lower() not in (
            "postgres",
            "postgresql",
        ):
            return self.run_queries(
                queries,
                pkDatabaseType,
                databaseType,
                metadata,
                runQueryConfig,
                overwrite_cache,
            )

        results = {}
        if not queries:
            return {"queryResults": []}

        def _normalize_db_type(db_type):
            if not db_type:
                return None
            lowered = db_type.lower()
            if lowered in ("postgresql", "postgres"):
                return "postgres"
            return lowered

        normalized_metadata_db = _normalize_db_type(databaseType)
        normalized_backend_db = _normalize_db_type(pkDatabaseType)

        def _is_read_only_query(query):
            if not isinstance(query, str):
                return False
            normalized = query.strip().lower()
            if not normalized:
                return False
            if not normalized.startswith(("select", "with", "show", "explain")):
                return False
            if "for update" in normalized or "for share" in normalized:
                return False
            if normalized.startswith("with") and re.search(
                r"\b(insert|update|delete|merge)\b", normalized
            ):
                return False
            return not re.search(
                r"\b(insert|update|delete|merge|alter|drop|create|truncate|grant|revoke|lock|vacuum|analyze|refresh)\b",
                normalized,
            )

        def _can_parallelize_query_batch(query_batch):
            return (
                normalized_backend_db == "postgres"
                and len(query_batch) > 1
                and all(_is_read_only_query(query) for query in query_batch)
            )

        def _get_max_concurrency():
            configured_max = (runQueryConfig or {}).get("maxConcurrency", 8)
            if not isinstance(configured_max, int) or configured_max < 1:
                configured_max = 8
            return min(len(queries), configured_max)

        async def _query_many(query_batch):
            semaphore = asyncio.Semaphore(_get_max_concurrency())

            async def _run(query):
                async with semaphore:
                    return await self.target_connection.query_async(
                        query, overwrite_cache
                    )

            return await asyncio.gather(*(_run(query) for query in query_batch))

        if (
            normalized_metadata_db
            and normalized_backend_db
            and normalized_metadata_db != normalized_backend_db
        ):
            return {"dbMismatched": True, "backendDatabaseType": pkDatabaseType}
        if runQueryConfig and runQueryConfig.get("arrayToMap"):
            mapped_array = array_to_map(
                queries,
                runQueryConfig.get("arrayToMap"),
                metadata,
                self.target_connection,
            )

            return {"queryResults": [], "mapped_array": mapped_array}
        elif runQueryConfig and runQueryConfig.get("getColumns"):
            query_results = await self.target_connection.query_async(
                queries[0].strip().rstrip(";") + " limit 1000",
                overwrite_cache,
            )
            results["columns"] = [
                {
                    "fieldType": convert_type_to_postgres(result["dataTypeID"]),
                    "name": result["name"],
                    "displayName": result["name"],
                    "isVisible": True,
                    "field": result["name"],
                }
                for result in query_results["fields"]
            ]
        elif runQueryConfig and runQueryConfig.get("getColumnsForSchema"):
            query_results = []
            for table in queries:
                if not table.get("viewQuery") or (
                    not table.get("isSelectStar") and not table.get("customFieldInfo")
                ):
                    query_results.append(table)
                    continue

                limit = ""
                if runQueryConfig.get("limitBy"):
                    limit = f" limit {runQueryConfig.get('limitBy')}"

                try:
                    query_result = await self.target_connection.query_async(
                        f"{table['viewQuery'].strip().rstrip(';')} {limit}",
                        overwrite_cache,
                    )
                    columns = [
                        {
                            "fieldType": convert_type_to_postgres(field["dataTypeID"]),
                            "name": field["name"],
                            "displayName": field["name"],
                            "isVisible": True,
                            "field": field["name"],
                        }
                        for field in query_result["fields"]
                    ]
                    query_results.append(
                        {**table, "columns": columns, "rows": query_result["rows"]}
                    )
                except Exception as e:
                    query_results.append(
                        {**table, "error": f"Error fetching columns {e}"}
                    )

            results["queryResults"] = query_results
            if runQueryConfig.get("fieldsToRemove"):
                results["queryResults"] = [
                    {
                        **table,
                        "columns": [
                            column
                            for column in table.get("columns", [])
                            if column["name"] not in runQueryConfig["fieldsToRemove"]
                        ],
                    }
                    for table in query_results
                ]
            return results
        elif runQueryConfig and runQueryConfig.get("getTables"):
            # Schema introspection path remains on the sync helper currently.
            tables = get_schema_tables_by_db(
                self.target_connection.database_type,
                self.target_connection.connection,
                runQueryConfig["schemaNames"],
            )
            schema = get_schema_column_info_by_db(
                self.target_connection.database_type,
                self.target_connection.connection,
                runQueryConfig["schemaNames"],
                tables,
            )
            results["queryResults"] = schema
        elif runQueryConfig and runQueryConfig.get("runIndividualQueries"):
            async def _run_individual_query(query):
                try:
                    run_query = query
                    if runQueryConfig.get("limitBy"):
                        run_query = self.apply_limit(query, runQueryConfig["limitBy"])
                    return await self.target_connection.query_async(
                        run_query, overwrite_cache
                    )
                except Exception as e:
                    return {
                        "query": query,
                        "error": str(e),
                    }

            if _can_parallelize_query_batch(queries):
                query_results = await asyncio.gather(
                    *(_run_individual_query(query) for query in queries)
                )
            else:
                query_results = []
                for query in queries:
                    query_results.append(await _run_individual_query(query))
            results["queryResults"] = query_results
        else:
            if runQueryConfig and runQueryConfig.get("limitThousand"):
                queries = [
                    query.strip().rstrip(";") + " limit 1000" for query in queries
                ]
            elif runQueryConfig and runQueryConfig.get("limitBy"):
                queries = [
                    query.strip().rstrip(";")
                    + f" limit {runQueryConfig.get('limitBy')}"
                    for query in queries
                ]
            if _can_parallelize_query_batch(queries):
                query_results = await _query_many(queries)
            else:
                query_results = []
                for query in queries:
                    query_results.append(
                        await self.target_connection.query_async(query, overwrite_cache)
                    )
            results["queryResults"] = query_results
            if runQueryConfig and runQueryConfig.get("fieldsToRemove"):
                results["queryResults"] = [
                    remove_fields(query_result, runQueryConfig.get("fieldsToRemove"))
                    for query_result in results["queryResults"]
                ]
            if runQueryConfig and runQueryConfig.get("convertDatatypes"):
                for query_result in results["queryResults"]:
                    query_result["fields"] = [
                        {
                            "name": field["name"],
                            "displayName": field["name"],
                            "field": field["name"],
                            "isVisible": True,
                            "dataTypeID": field["dataTypeID"],
                            "fieldType": convert_type_to_postgres(field["dataTypeID"]),
                        }
                        for field in query_result["fields"]
                    ]

        return results

    def run_queries(
        self,
        queries,
        pkDatabaseType,
        databaseType=None,
        metadata=None,
        runQueryConfig=None,
        overwrite_cache=False,
    ):
        results = {}
        if not queries:
            return {"queryResults": []}
        def _normalize_db_type(db_type):
            if not db_type:
                return None
            lowered = db_type.lower()
            if lowered in ("postgresql", "postgres"):
                return "postgres"
            return lowered

        normalized_metadata_db = _normalize_db_type(databaseType)
        normalized_backend_db = _normalize_db_type(pkDatabaseType)

        def _is_read_only_query(query):
            if not isinstance(query, str):
                return False
            normalized = query.strip().lower()
            if not normalized:
                return False
            if not normalized.startswith(("select", "with", "show", "explain")):
                return False
            if "for update" in normalized or "for share" in normalized:
                return False
            if normalized.startswith("with") and re.search(
                r"\b(insert|update|delete|merge)\b", normalized
            ):
                return False
            return not re.search(
                r"\b(insert|update|delete|merge|alter|drop|create|truncate|grant|revoke|lock|vacuum|analyze|refresh)\b",
                normalized,
            )

        def _can_parallelize_query_batch(query_batch):
            return (
                normalized_backend_db == "postgres"
                and len(query_batch) > 1
                and all(_is_read_only_query(query) for query in query_batch)
            )

        def _get_max_concurrency():
            configured_max = (runQueryConfig or {}).get("maxConcurrency", 8)
            if not isinstance(configured_max, int) or configured_max < 1:
                configured_max = 8
            return min(len(queries), configured_max)

        def _parallel_query_batch(query_batch):
            worker_count = _get_max_concurrency()
            with ThreadPoolExecutor(max_workers=worker_count) as executor:
                return list(
                    executor.map(
                        lambda query: self.target_connection.query(query, overwrite_cache),
                        query_batch,
                    )
                )

        if (
            normalized_metadata_db
            and normalized_backend_db
            and normalized_metadata_db != normalized_backend_db
        ):
            return {"dbMismatched": True, "backendDatabaseType": pkDatabaseType}
        if runQueryConfig and runQueryConfig.get("arrayToMap"):
            mapped_array = array_to_map(
                queries,
                runQueryConfig.get("arrayToMap"),
                metadata,
                self.target_connection,
            )

            return {"queryResults": [], "mapped_array": mapped_array}
        elif runQueryConfig and runQueryConfig.get("getColumns"):
            query_results = self.target_connection.query(
                queries[0].strip().rstrip(";") + " limit 1000",
                overwrite_cache,
            )
            results["columns"] = [
                {
                    "fieldType": convert_type_to_postgres(result["dataTypeID"]),
                    "name": result["name"],
                    "displayName": result["name"],
                    "isVisible": True,
                    "field": result["name"],
                }
                for result in query_results["fields"]
            ]
        elif runQueryConfig and runQueryConfig.get("getColumnsForSchema"):
            query_results = []
            for table in queries:
                if not table.get("viewQuery") or (
                    not table.get("isSelectStar") and not table.get("customFieldInfo")
                ):
                    query_results.append(table)
                    continue

                limit = ""
                if runQueryConfig.get("limitBy"):
                    limit = f" limit {runQueryConfig.get('limitBy')}"

                try:
                    query_result = self.target_connection.query(
                        f"{table['viewQuery'].strip().rstrip(';')} {limit}",
                        overwrite_cache,
                    )
                    columns = [
                        {
                            "fieldType": convert_type_to_postgres(field["dataTypeID"]),
                            "name": field["name"],
                            "displayName": field["name"],
                            "isVisible": True,
                            "field": field["name"],
                        }
                        for field in query_result["fields"]
                    ]
                    query_results.append(
                        {**table, "columns": columns, "rows": query_result["rows"]}
                    )
                except Exception as e:
                    query_results.append(
                        {**table, "error": f"Error fetching columns {e}"}
                    )

            results["queryResults"] = query_results
            if runQueryConfig.get("fieldsToRemove"):
                results["queryResults"] = [
                    {
                        **table,
                        "columns": [
                            column
                            for column in table.get("columns", [])
                            if column["name"] not in runQueryConfig["fieldsToRemove"]
                        ],
                    }
                    for table in query_results
                ]
            return results
        elif runQueryConfig and runQueryConfig.get("getTables"):
            tables = get_schema_tables_by_db(
                self.target_connection.database_type,
                self.target_connection.connection,
                runQueryConfig["schemaNames"],
            )
            schema = get_schema_column_info_by_db(
                self.target_connection.database_type,
                self.target_connection.connection,
                runQueryConfig["schemaNames"],
                tables,
            )
            results["queryResults"] = schema
        elif runQueryConfig and runQueryConfig.get("runIndividualQueries"):
            # so that one query doesn't fail the whole thing
            # the only reason this isn't the default behavior is for backwards compatibility
            def _run_individual_query(query):
                try:
                    run_query = query
                    if runQueryConfig.get("limitBy"):
                        run_query = self.apply_limit(query, runQueryConfig["limitBy"])
                    return self.target_connection.query(run_query, overwrite_cache)
                except Exception as e:
                    return {
                        "query": query,
                        "error": str(e),
                    }

            if _can_parallelize_query_batch(queries):
                worker_count = _get_max_concurrency()
                with ThreadPoolExecutor(max_workers=worker_count) as executor:
                    query_results = list(executor.map(_run_individual_query, queries))
            else:
                query_results = [_run_individual_query(query) for query in queries]
            results["queryResults"] = query_results
        else:
            if runQueryConfig and runQueryConfig.get("limitThousand"):
                queries = [
                    query.strip().rstrip(";") + " limit 1000" for query in queries
                ]
            elif runQueryConfig and runQueryConfig.get("limitBy"):
                queries = [
                    query.strip().rstrip(";")
                    + f" limit {runQueryConfig.get('limitBy')}"
                    for query in queries
                ]
            if _can_parallelize_query_batch(queries):
                query_results = _parallel_query_batch(queries)
            else:
                query_results = [
                    self.target_connection.query(query, overwrite_cache)
                    for query in queries
                ]
            results["queryResults"] = query_results
            if runQueryConfig and runQueryConfig.get("fieldsToRemove"):
                results["queryResults"] = [
                    remove_fields(query_result, runQueryConfig.get("fieldsToRemove"))
                    for query_result in results["queryResults"]
                ]
            if runQueryConfig and runQueryConfig.get("convertDatatypes"):
                for query_result in results["queryResults"]:
                    query_result["fields"] = [
                        {
                            "name": field["name"],
                            "displayName": field["name"],
                            "field": field["name"],
                            "isVisible": True,
                            "dataTypeID": field["dataTypeID"],
                            "fieldType": convert_type_to_postgres(field["dataTypeID"]),
                        }
                        for field in query_result["fields"]
                    ]

        return results
    
    async def post_quill(self, path, payload, use_cache=True):
        # Custom JSON Encoder to handle Enums
        class EnumEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, Enum):
                    return obj.value  # Convert enum to its value (string in this case)
                return super().default(obj)

        cache = getattr(self.target_connection, "cache", None)
        # Match Node SDK behavior: skip post cache for tenant-mapped-flags and set-broken-view.
        can_use_cache = bool(
            use_cache and cache and path not in ("tenant-mapped-flags", "set-broken-view")
        )
        cache_key = None
        if can_use_cache:
            remove_cache_override = dict(payload or {})
            remove_cache_override.pop("overwriteCache", None)
            key_payload = json.dumps(
                remove_cache_override,
                cls=EnumEncoder,
                separators=(",", ":"),
                ensure_ascii=False,
            )
            hash_value = hashlib.sha256(key_payload.encode("utf-8")).hexdigest()
            cache_key = f"quill:post:{path}:{hash_value}"
            if not (payload or {}).get("overwriteCache"):
                try:
                    cached = cache.get(cache_key)
                    if cached:
                        try:
                            return json.loads(cached)
                        except Exception:
                            try:
                                cache.delete(cache_key)
                            except Exception:
                                pass
                except Exception:
                    pass

        url = f"{self.baseUrl}/sdk/{path}"
        # Set content type to application/json
        headers = {"Authorization": f"Bearer {self.private_key}", "Content-Type": "application/json"}
        encoded = json.dumps(payload, cls=EnumEncoder)
        response = await asyncio.to_thread(
            requests.post,
            url,
            data=encoded,
            headers=headers,
        )
        response_json = response.json()

        if can_use_cache and cache_key and not response_json.get("error"):
            ttl = getattr(self.target_connection, "ttl", None)
            try:
                if ttl:
                    cache.set(cache_key, json.dumps(response_json), "EX", ttl)
                else:
                    cache.set(cache_key, json.dumps(response_json))
            except Exception:
                pass

        return response_json



"""
Pivot Template System - SDK SIDE

This file contains all the logic needed on the SDK to:
1. Hydrate pivot query templates with actual distinct values
2. Parse distinct values from different database result formats
3. Validate templates before hydration

This runs on the customer's Python SDK where customer data is accessible.
Takes templates from server and populates them with actual data.
"""

import json
import re
from typing import List, Dict, Any, Optional, TypedDict

# Constants
MAX_PIVOT_UNIQUE_VALUES = 250
PIVOT_COLUMN_MARKER = "{{QUILL_PIVOT_COLUMNS}}"
PIVOT_COLUMN_ALIAS_MARKER = "{{QUILL_PIVOT_COLUMN_ALIASES}}"


# Types
class PivotAggregation(TypedDict, total=False):
    aggregationType: str
    valueField: Optional[str]
    valueFieldType: Optional[str]
    valueField2: Optional[str]
    valueField2Type: Optional[str]


class PivotConfig(TypedDict, total=False):
    requiresDistinctValues: bool
    columnField: Optional[str]
    rowField: Optional[str]
    rowFieldType: Optional[str]
    aggregations: List[PivotAggregation]
    databaseType: str
    dateBucket: Optional[str]
    pivotType: str
    sort: Optional[bool]
    sortField: Optional[str]
    sortDirection: Optional[str]
    rowLimit: Optional[int]


# ============================================================================
# HELPER FUNCTIONS
# ============================================================================


def process_single_quotes(value: str, database_type: str) -> str:
    """Process single quotes based on database type."""
    if database_type.lower() in ["postgresql", "snowflake", "clickhouse"]:
        return value.replace("'", "''")
    return value.replace("'", "\\'")


def process_agg_type(agg_type: str, has_column_field: bool = False) -> str:
    """Process aggregation type."""
    if agg_type == "count" and has_column_field:
        return "SUM"
    return "AVG" if agg_type and agg_type.lower() == "average" else (agg_type.lower() if agg_type else "")


def replace_bigquery_special_characters(column: str) -> str:
    """Replace BigQuery special characters."""
    return column.replace("/", "quill_forward_slash")


def process_column_reference(
    column: str,
    database_type: str,
    fallback_on_null: Optional[str] = None,
    is_column_field_alias: bool = False,
    is_value_field_alias: bool = False
) -> str:
    """Process column reference based on database type."""
    db = database_type.lower()
    
    if db in ["postgresql", "clickhouse"]:
        if column == "":
            return f'"{fallback_on_null}"' if fallback_on_null else '"_"'
        if is_column_field_alias:
            return f'"{column.replace(chr(34), "")}"'
        column_parts = column.split(".")
        if len(column_parts) > 1:
            return '"' + '","'.join([part.replace('"', '') for part in column_parts]) + '"'
        return f'"{column.replace(chr(34), "")}"'
    
    elif db == "mysql":
        if column == "":
            return fallback_on_null if fallback_on_null else "_"
        if is_column_field_alias:
            return f"`{column.replace('`', '').replace(chr(34), '')}`"
        column_parts = column.split(".")
        if len(column_parts) > 1:
            return "`" + "`.`".join([part.replace("`", "") for part in column_parts]) + "`"
        return f"`{column.replace('`', '')}`"
    
    elif db == "snowflake":
        if column == "":
            return fallback_on_null if fallback_on_null else "_"
        if is_column_field_alias:
            return f'"{column.replace(chr(34), "")}"'
        if is_value_field_alias:
            cleaned_column = column.replace(")", "").replace("(", "_")
            return cleaned_column
        return column
    
    elif db == "bigquery":
        if column == "":
            return f"`{fallback_on_null}`" if fallback_on_null else "`_`"
        if is_column_field_alias:
            return f"`{replace_bigquery_special_characters(column)}`"
        column_parts = column.split(".")
        if len(column_parts) > 1:
            return "`" + "`.`".join([part for part in column_parts]) + "`"
        return f"`{column}`"
    
    elif db == "mssql":
        if column == "":
            return f"[{fallback_on_null}]" if fallback_on_null else "[_]"
        if is_column_field_alias:
            return f"[{column}]"
        column_parts = column.split(".")
        if len(column_parts) > 1:
            return "[" + "].[".join([part for part in column_parts]) + "]"
        return f"[{column}]"
    
    elif db == "databricks":
        if column == "":
            return f"`{fallback_on_null}`" if fallback_on_null else "`_`"
        if is_column_field_alias:
            return f"`{column}`"
        column_parts = column.split(".")
        if len(column_parts) > 1:
            return "`" + "`.`".join([part for part in column_parts]) + "`"
        return f"`{column}`"
    
    else:
        return column


def process_value_field(agg_type: str, database_type: str, value_field: str) -> str:
    """Process value field based on aggregation type."""
    if agg_type in ["min", "max"] or (agg_type and agg_type.lower() == "average"):
        return f"{process_column_reference(value_field, database_type)} ELSE null"
    if agg_type == "count":
        return "1 ELSE 0"
    return f"{process_column_reference(value_field, database_type)} ELSE 0" if value_field else "1 ELSE 0"


# ============================================================================
# DISTINCT VALUES PARSING
# ============================================================================


def parse_distinct_values(query_result: Dict[str, Any], database_type: str) -> List[str]:
    """
    Parses distinct values from database query results.
    Different databases return different formats.
    """
    if not query_result or not query_result.get("rows") or len(query_result["rows"]) == 0:
        return []
    
    row = query_result["rows"][0]
    distinct_values = []
    
    db = database_type.lower()
    
    if db in ["postgresql", "bigquery", "snowflake", "databricks", "clickhouse"]:
        # These return arrays in string_values field
        if "string_values" in row:
            if isinstance(row["string_values"], list):
                distinct_values = row["string_values"]
            elif isinstance(row["string_values"], str):
                # Handle JSON string arrays
                try:
                    distinct_values = json.loads(row["string_values"])
                except:
                    distinct_values = []
    
    elif db == "mysql":
        # MySQL returns JSON_ARRAYAGG which should be an array
        if "string_values" in row:
            if isinstance(row["string_values"], list):
                distinct_values = row["string_values"]
            elif isinstance(row["string_values"], str):
                try:
                    distinct_values = json.loads(row["string_values"])
                except:
                    distinct_values = []
    
    elif db == "mssql":
        # MS SQL returns comma-separated string
        if "string_values" in row and isinstance(row["string_values"], str):
            distinct_values = [v.strip() for v in row["string_values"].split(",")]
    
    else:
        distinct_values = []
    
    # Filter out null/undefined/empty values
    return [value for value in distinct_values if value is not None and value != ""]


# ============================================================================
# MATCH CASING FUNCTION
# ============================================================================


def match_casing(text: Optional[str], template: Optional[str]) -> str:
    """Matches the casing of text to template."""
    if not text or not template:
        return text or ""
    
    # Detect patterns
    def is_title_case(s: str) -> bool:
        return bool(re.match(r'^[A-Z][a-z]*([A-Z][a-z]*)*$', s))
    
    def is_camel_case(s: str) -> bool:
        return bool(re.match(r'^[a-z]+([A-Z][a-z]*)*$', s))
    
    def is_snake_case(s: str) -> bool:
        return bool(re.match(r'^[a-z0-9]+(_[a-z0-9]+)*$', s))
    
    def is_all_lower_case(s: str) -> bool:
        return bool(re.match(r'^[a-z]+$', s))
    
    def is_all_upper_case(s: str) -> bool:
        return bool(re.match(r'^[A-Z]+$', s))
    
    def is_capitalized(s: str) -> bool:
        return bool(re.match(r'^[A-Z][a-z]*$', s))
    
    def is_screaming_snake_case(s: str) -> bool:
        return bool(re.match(r'^[A-Z][A-Z0-9]*(_[A-Z0-9]+)*$', s))
    
    # Convert functions
    def to_title_case(s: str) -> str:
        return ''.join(word.capitalize() for word in re.split(r'[_\s]+', s.lower()))
    
    def to_camel_case(s: str) -> str:
        return re.sub(r'_(.)', lambda m: m.group(1).upper(), s.lower())
    
    def to_snake_case(s: str) -> str:
        return re.sub(r'[A-Z]', lambda m: f'_{m.group(0).lower()}', s)
    
    def to_screaming_snake_case(s: str) -> str:
        result = re.sub(r'([A-Z])', r'_\1', s)
        result = result.lstrip('_')
        return result.upper()
    
    # Match casing
    if is_title_case(template):
        return to_title_case(text)
    elif is_camel_case(template):
        return to_camel_case(text)
    elif is_snake_case(template):
        return to_snake_case(text)
    elif is_all_lower_case(template):
        return text.lower()
    elif is_all_upper_case(template):
        return text.upper()
    elif is_capitalized(template):
        return text.capitalize()
    elif is_screaming_snake_case(template):
        return to_screaming_snake_case(text)
    else:
        return text  # Default case if no specific pattern is detected


# ============================================================================
# TEMPLATE HYDRATION
# ============================================================================


def hydrate_pivot_template(
    template: str,
    distinct_values: List[str],
    config: PivotConfig
) -> str:
    """
    Hydrates a pivot query template with actual distinct values.
    This function should be called in the Python SDK after fetching distinct values.
    
    Args:
        template: The SQL template string containing markers
        distinct_values: Array of distinct values fetched from the database
        config: config about the pivot configuration
    
    Returns:
        Hydrated SQL query string ready to execute
    """
    column_field = config.get("columnField")
    row_field = config.get("rowField")
    aggregations = config.get("aggregations", [])
    database_type = config.get("databaseType", "postgresql")
    
    # If this pivot doesn't require distinct values, return as-is
    if not config.get("requiresDistinctValues") or not column_field or not row_field:
        return template
    
    # Filter and limit distinct values
    filtered_values = [
        value for value in distinct_values
        if value is not None and value != ""
    ][:MAX_PIVOT_UNIQUE_VALUES]
    
    # Get properly quoted column references
    column_field_alias = process_column_reference(
        column_field,
        database_type,
        None,
        False,
        True
    )
    
    row_field_alias = process_column_reference(
        row_field,
        database_type,
        None,
        False,
        True
    )
    
    # Generate column aliases for SELECT in quill_alias CTE
    column_aliases = []
    column_aliases.append(
        f"{process_column_reference(row_field, database_type, None, True)} AS {row_field_alias}"
    )
    
    # Generate CASE WHEN columns for each aggregation
    case_when_columns = []
    seen_aggs: Dict[str, Dict[str, int]] = {}
    
    for current_agg in aggregations:
        agg_type = current_agg.get("aggregationType", "")
        value_field = current_agg.get("valueField", "")
        
        # Track duplicate aggregation combos for disambiguation
        if agg_type in seen_aggs and value_field in seen_aggs[agg_type]:
            seen_aggs[agg_type][value_field] += 1
        else:
            if agg_type not in seen_aggs:
                seen_aggs[agg_type] = {}
            seen_aggs[agg_type][value_field] = 1
        
        disambiguation_index = str(seen_aggs[agg_type][value_field])
        if disambiguation_index == "1":
            disambiguation_index = ""
        
        value_field_alias = process_column_reference(
            current_agg.get("valueField") or row_field or "count",
            database_type,
            None,
            False,
            True
        )
        
        value_alias_substring = ""
        if current_agg.get("valueField"):
            value_alias_substring = f"{process_column_reference(current_agg['valueField'], database_type, None, True)} AS {value_field_alias}"
        
        # Handle disambiguation for multiple aggregations
        total_seen = sum(seen_aggs[agg_type].values())
        disambiguation_field = ""
        if total_seen > 1:
            disambiguation_field = f"_{current_agg.get('valueField', '')}{disambiguation_index}"
        
        disambiguation = ""
        if len(aggregations) > 1:
            if disambiguation_field:
                disambiguation = f"{disambiguation_field}_{match_casing(agg_type, current_agg.get('valueField'))}"
            else:
                disambiguation = f"_{agg_type}"
        
        # Wrap boolean fields in CASE WHEN
        value_expr = ""
        if current_agg.get("valueFieldType") == "bool":
            value_expr = f"CASE WHEN {value_field_alias} THEN 1 ELSE 0 END"
        else:
            value_expr = process_value_field(
                agg_type,
                database_type,
                value_field_alias
            )
        
        # Handle percentage aggregations specially
        if agg_type == "percentage":
            value_field2 = current_agg.get("valueField2") or current_agg.get("valueField") or "count"
            value_field2_alias = process_column_reference(
                value_field2,
                database_type,
                None,
                False,
                True
            )
            
            value_field2_type = current_agg.get("valueField2Type") or current_agg.get("valueFieldType")
            value2_expr = ""
            if value_field2_type == "bool":
                value2_expr = f"CASE WHEN {value_field2_alias} THEN 1 ELSE 0 END"
            else:
                value2_expr = value_field2_alias
            
            value2_alias_substring = ""
            if current_agg.get("valueField2") and current_agg.get("valueField") != current_agg.get("valueField2"):
                value2_alias_substring = f"{process_column_reference(current_agg['valueField2'], database_type, None, True)} AS {value_field2_alias}"
            
            # Percentage with same field for numerator and denominator
            if current_agg.get("valueField") == current_agg.get("valueField2") or not current_agg.get("valueField2"):
                for column in filtered_values:
                    case_when_columns.append(
                        f"CAST(sum(CASE WHEN {column_field_alias} = '{process_single_quotes(column, database_type)}' THEN {value_expr} END) AS FLOAT) / GREATEST(sum({value2_expr}), 1) AS {process_column_reference(column + disambiguation, database_type, '_', True)}"
                    )
            else:
                # Percentage with different fields
                for column in filtered_values:
                    case_when_columns.append(
                        f"CAST(sum(CASE WHEN {column_field_alias} = '{process_single_quotes(column, database_type)}' THEN {value_expr} END) AS FLOAT) / GREATEST(sum(CASE WHEN {column_field_alias} = '{process_single_quotes(column, database_type)}' THEN {value2_expr} END), 1) AS {process_column_reference(column + disambiguation, database_type, '_', True)}"
                    )
                if value2_alias_substring:
                    column_aliases.append(value2_alias_substring)
        else:
            # Standard aggregations (sum, count, avg, min, max)
            for column in filtered_values:
                case_when_columns.append(
                    f"{process_agg_type(agg_type, True)}(CASE WHEN {column_field_alias} = '{process_single_quotes(column, database_type)}' THEN {value_expr} END) AS {process_column_reference(column + disambiguation, database_type, '_', True)}"
                )
        
        if value_alias_substring:
            column_aliases.append(value_alias_substring)
    
    # Add the column field to the aliases
    column_aliases.append(
        f"{process_column_reference(column_field, database_type, None, True)} AS {column_field_alias}"
    )
    
    # Remove duplicates
    unique_column_aliases = list(dict.fromkeys(column_aliases))
    
    # Replace markers with actual SQL
    hydrated_template = template.replace(
        PIVOT_COLUMN_ALIAS_MARKER,
        ", ".join(unique_column_aliases)
    ).replace(
        PIVOT_COLUMN_MARKER,
        ", ".join(case_when_columns)
    )
    
    return hydrated_template


# ============================================================================
# VALIDATION
# ============================================================================


def validate_template(template: str, config: PivotConfig) -> Dict[str, Any]:
    """Validates that a template can be hydrated with the given config."""
    errors = []
    
    if not template:
        errors.append("Template is empty")
    
    if config.get("requiresDistinctValues"):
        if PIVOT_COLUMN_MARKER not in template:
            errors.append(f"Template is missing {PIVOT_COLUMN_MARKER} marker")
        if PIVOT_COLUMN_ALIAS_MARKER not in template:
            errors.append(f"Template is missing {PIVOT_COLUMN_ALIAS_MARKER} marker")
        if not config.get("columnField"):
            errors.append("config is missing columnField")
        if not config.get("rowField"):
            errors.append("config is missing rowField")
    
    if not config.get("aggregations") or len(config.get("aggregations", [])) == 0:
        errors.append("config is missing aggregations")
    
    return {
        "valid": len(errors) == 0,
        "errors": errors
    }





# __init__.py

from .run_query_processes import remove_fields, array_to_map
from .filters import Filter, FilterType, FieldType, StringOperator, NumberOperator, NullOperator, DateOperator, convert_custom_filter
from .post_quill_executor import ParallelPostQuill



import os
import importlib
from quillsql.assets.pgtypes import PG_TYPES

DEFAULT_POOL_MIN_SIZE = 0
DEFAULT_POOL_MAX_SIZE = 10
# Effective default from pg-pool constructor when idleTimeoutMillis
# is not explicitly provided to new Pool(...): 10000ms (10 seconds).
DEFAULT_POOL_MAX_IDLE_SECONDS = 10.0
# node-postgres default maxLifetimeSeconds is 0 (disabled).
DEFAULT_POOL_MAX_LIFETIME_SECONDS = 0.0


def format_postgres(connection_string):
    return connection_string


def _env_int(name, default, minimum=0):
    raw = os.getenv(name)
    if raw is None or str(raw).strip() == "":
        return default
    try:
        parsed = int(raw)
    except (TypeError, ValueError):
        return default
    return max(minimum, parsed)


def _env_float(name, default, minimum=0.0):
    raw = os.getenv(name)
    if raw is None or str(raw).strip() == "":
        return default
    try:
        parsed = float(raw)
    except (TypeError, ValueError):
        return default
    return max(minimum, parsed)


def _pool_kwargs():
    min_size = _env_int("QUILL_PG_POOL_MIN_SIZE", DEFAULT_POOL_MIN_SIZE, minimum=0)
    max_size = _env_int("QUILL_PG_POOL_MAX_SIZE", DEFAULT_POOL_MAX_SIZE, minimum=1)
    if max_size < min_size:
        max_size = min_size
    return {
        "min_size": min_size,
        "max_size": max_size,
        "max_idle": _env_float(
            "QUILL_PG_POOL_MAX_IDLE_SECONDS",
            DEFAULT_POOL_MAX_IDLE_SECONDS,
            minimum=0.0,
        ),
        "max_lifetime": _env_float(
            "QUILL_PG_POOL_MAX_LIFETIME_SECONDS",
            DEFAULT_POOL_MAX_LIFETIME_SECONDS,
            minimum=0.0,
        ),
    }


def _build_conninfo(config, using_connection_string):
    if using_connection_string:
        return config
    return (
        f"dbname={config['dbname']} "
        f"user={config['user']} "
        f"password={config['password']} "
        f"host={config['host']} "
        f"port={config['port']}"
    )


def connect_to_postgres(config, usingConnectionString):
    psycopg_pool = importlib.import_module("psycopg_pool")
    psycopg_rows = importlib.import_module("psycopg.rows")
    connection_pool_cls = psycopg_pool.ConnectionPool
    os.environ["PGGSSENCMODE"] = "disable"
    conn_kwargs = {"autocommit": True, "row_factory": psycopg_rows.dict_row}
    conninfo = _build_conninfo(config, usingConnectionString)
    pool_kwargs = _pool_kwargs()

    return connection_pool_cls(
        conninfo=conninfo,
        kwargs=conn_kwargs,
        open=True,
        **pool_kwargs,
    )


def connect_to_postgres_async(config, usingConnectionString):
    psycopg_pool = importlib.import_module("psycopg_pool")
    psycopg_rows = importlib.import_module("psycopg.rows")
    connection_pool_cls = psycopg_pool.AsyncConnectionPool
    os.environ["PGGSSENCMODE"] = "disable"
    conn_kwargs = {"autocommit": True, "row_factory": psycopg_rows.dict_row}
    conninfo = _build_conninfo(config, usingConnectionString)
    pool_kwargs = _pool_kwargs()

    # Use explicit open=False to align with psycopg_pool guidance.
    return connection_pool_cls(
        conninfo=conninfo,
        kwargs=conn_kwargs,
        open=False,
        **pool_kwargs,
    )


def run_query_postgres(query, pool):
    with pool.connection() as connection:
        with connection.cursor() as cursor:
            cursor.execute(query)
            result = cursor.fetchall()
            fields = [
                {"name": desc[0], "dataTypeID": desc[1]} for desc in cursor.description
            ]
    return {"rows": result, "fields": fields}


async def run_query_postgres_async(query, pool):
    async with pool.connection() as connection:
        async with connection.cursor() as cursor:
            await cursor.execute(query)
            result = await cursor.fetchall()
            fields = [
                {"name": desc[0], "dataTypeID": desc[1]} for desc in cursor.description
            ]
    return {"rows": result, "fields": fields}


def disconnect_from_postgres(pool):
    if pool is None:
        return
    if getattr(pool, "closed", False):
        return

    try:
        # Per psycopg_pool docs, close(timeout=...) controls thread wait.
        # timeout=0 avoids waiting/joining worker threads from cleanup paths
        # where current-thread joins can surface.
        pool.close(timeout=0)
    except TypeError:
        # Compatibility fallback for older pool signatures.
        pool.close()
    return


async def disconnect_from_postgres_async(pool):
    if pool is None:
        return
    if getattr(pool, "closed", False):
        return

    try:
        await pool.close(timeout=0)
    except TypeError:
        await pool.close()
    return


# getTablesBySchemaPostgres


def get_tables_by_schema_postgres(connection, schema_names):
    all_tables = []
    for schema_name in schema_names:
        query = f"""
            SELECT table_name, table_schema
            FROM information_schema.tables
            WHERE table_schema = '{schema_name}'

            UNION

            SELECT c.relname as table_name, n.nspname as table_schema
            FROM pg_class c
            JOIN pg_namespace n ON c.relnamespace = n.oid
            WHERE n.nspname = '{schema_name}' 
            AND c.relkind = 'm';
            """
        results = run_query_postgres(query, connection)
        for row in results["rows"]:
            cur_table = {}
            cur_table["table_name"] = row["table_name"]
            cur_table["schema_name"] = row["table_schema"]
            all_tables.append(cur_table)
    return all_tables


# getSchemaColumnInfoPostgres


def get_schema_column_info_postgres(connection, schema_name, table_names):
    all_columns = []
    for table_name in table_names:
        query = f"""
            SELECT column_name as "column_name", udt_name as "field_type", ordinal_position as "sort_number"
            FROM information_schema.columns
            WHERE table_schema = '{table_name['schema_name']}' 
            AND table_name = '{table_name['table_name']}'
            
            UNION

            SELECT a.attname as "column_name", t.typname as "field_type", a.attnum as "sort_number"
            FROM pg_attribute a
            JOIN pg_class c ON a.attrelid = c.oid
            JOIN pg_namespace n ON c.relnamespace = n.oid
            JOIN pg_type t ON a.atttypid = t.oid
            WHERE n.nspname = '{table_name['schema_name']}'
                AND c.relname = '{table_name['table_name']}'
                AND c.relkind = 'm'
                AND a.attnum > 0
                AND NOT a.attisdropped
            ORDER BY "sort_number"
            """
        results = run_query_postgres(query, connection)
        columns = []
        for row in results["rows"]:
            pg_type = next(
                (
                    pg_type
                    for pg_type in PG_TYPES
                    if pg_type["typname"] == row["field_type"]
                ),
                None,
            )
            if pg_type is None:
                pg_type = 1043
            columns.append(
                {
                    "columnName": row["column_name"],
                    "displayName": row["column_name"],
                    "dataTypeID": pg_type["oid"],
                    "fieldType": row["field_type"],
                }
            )
        all_columns.append(
            {
                "tableName": table_name["schema_name"] + "." + table_name["table_name"],
                "displayName": table_name["schema_name"]
                + "."
                + table_name["table_name"],
                "columns": columns,
            }
        )
    return all_columns



import os
import sys
import asyncio
from unittest.mock import MagicMock, patch

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from quillsql.core import Quill


@patch("quillsql.core.CachedConnection")
def test_query_handles_none_tenant_flag_responses(mock_cached_connection):
    mock_connection = MagicMock()
    mock_connection.database_type = "PostgreSQL"
    mock_cached_connection.return_value = mock_connection

    quill = Quill(
        private_key=os.getenv("PRIVATE_KEY"),
        database_type=os.getenv("DB_TYPE"),
        database_connection_string=os.getenv("DB_URL")
    )

    quill.run_queries = MagicMock(
        side_effect=[
            {"queryResults": [None]},
            {"queryResults": []},
        ]
    )

    def post_quill_side_effect(path, payload):
        if path == "tenant-mapped-flags":
            return {
                "queries": [
                    'SELECT * FROM (SELECT "id" AS "customer_id", "name" AS "quill_label", "id" AS "quill_flag" FROM "public"."customers") AS "subq_owner_query" WHERE "customer_id" IN (2)'
                ],
                "metadata": {"queryOrder": ["customer_id"]},
            }
        return {"queries": [], "metadata": {}}

    quill.post_quill = MagicMock(side_effect=post_quill_side_effect)

    result = asyncio.run(
        quill.query(
            tenants=["tenant-1"],
            metadata={
                "task": "report",
                "clientId": "client-123",
                "reportId": "report-456",
                "databaseType": "PostgreSQL",
            },
        )
    )

    assert result["status"] == "success"
    assert result["data"] == {}


@patch("quillsql.core.CachedConnection")
def test_query_handles_tenant_flag_metadata_none(mock_cached_connection):
    mock_connection = MagicMock()
    mock_connection.database_type = "PostgreSQL"
    mock_cached_connection.return_value = mock_connection

    quill = Quill(
       private_key=os.getenv("PRIVATE_KEY"),
        database_type=os.getenv("DB_TYPE"),
        database_connection_string=os.getenv("DB_URL")
    )

    quill.run_queries = MagicMock(
        side_effect=[
            {"queryResults": [{"rows": [{"quill_flag": "alpha"}]}]},
            {"queryResults": []},
        ]
    )

    def post_quill_side_effect(path, payload):
        if path == "tenant-mapped-flags":
            return {
                "queries": [
                    'SELECT * FROM (SELECT "id" AS "customer_id", "name" AS "quill_label", "id" AS "quill_flag" FROM "public"."customers") AS "subq_owner_query" WHERE "customer_id" IN (2)'
                ],
                "metadata": None,
            }
        return {"queries": [], "metadata": {}}

    quill.post_quill = MagicMock(side_effect=post_quill_side_effect)

    result = asyncio.run(
        quill.query(
            tenants=[{"tenantField": "customer_id", "tenantIds": [2]}],
            metadata={
                "task": "report",
                "clientId": "client-123",
                "reportId": "report-456",
                "databaseType": "PostgreSQL",
            },
        )
    )

    assert result["status"] == "success"
    assert result["data"] == {}


@patch("quillsql.core.CachedConnection")
def test_query_handles_array_to_map_with_null_mapped_rows(mock_cached_connection):
    mock_connection = MagicMock()
    mock_connection.database_type = "PostgreSQL"
    mock_cached_connection.return_value = mock_connection

    quill = Quill(
        private_key=os.getenv("PRIVATE_KEY"),
        database_type=os.getenv("DB_TYPE"),
        database_connection_string=os.getenv("DB_URL")
    )

    quill.run_queries = MagicMock(
        side_effect=[
            {"queryResults": []},  # tenant-mapped-flags
            {"queryResults": [], "mapped_array": [None]},  # main query
        ]
    )

    def post_quill_side_effect(path, payload):
        if path == "tenant-mapped-flags":
            return {"queries": [], "metadata": {"queryOrder": []}}
        if path == "filter-options":
            return {
                "queries": ["SELECT 1"],
                "metadata": {
                    "filters": [{"label": "Driver"}],
                    "runQueryConfig": {
                        "arrayToMap": {"arrayName": "filters", "field": "options"}
                    },
                },
            }
        return {"queries": [], "metadata": {}}

    quill.post_quill = MagicMock(side_effect=post_quill_side_effect)

    result = asyncio.run(
        quill.query(
            tenants=[{"tenantField": "customer_id", "tenantIds": [2]}],
            metadata={
                "task": "filter-options",
                "clientId": "client-123",
                "filter": {"field": "user_name"},
            },
            flags=[],
        )
    )

    assert result["status"] == "success"
    assert result["data"]["filters"][0]["options"] == []


@patch("quillsql.core.CachedConnection")
def test_query_handles_fields_to_remove_with_null_results(mock_cached_connection):
    mock_connection = MagicMock()
    mock_connection.database_type = "PostgreSQL"
    mock_cached_connection.return_value = mock_connection

    quill = Quill(
        private_key=os.getenv("PRIVATE_KEY"),
        database_type=os.getenv("DB_TYPE"),
        database_connection_string=os.getenv("DB_URL")
    )

    quill.run_queries = MagicMock(
        side_effect=[
            {"queryResults": []},  # tenant-mapped-flags
            {"queryResults": [None]},  # main run_queries
        ]
    )

    def post_quill_side_effect(path, payload):
        if path == "tenant-mapped-flags":
            return {"queries": [], "metadata": {"queryOrder": []}}
        if path == "report":
            return {
                "queries": ["SELECT 1"],
                "metadata": {
                    "runQueryConfig": {"fieldsToRemove": ["customer_id"]},
                },
            }
        return {"queries": [], "metadata": {}}

    quill.post_quill = MagicMock(side_effect=post_quill_side_effect)

    result = asyncio.run(
        quill.query(
            tenants=[{"tenantField": "customer_id", "tenantIds": [2]}],
            metadata={
                "task": "report",
                "clientId": "client-123",
                "reportId": "report-456",
                "databaseType": "PostgreSQL",
            },
            flags=[],
        )
    )

    assert result["status"] == "success"
    assert result["queries"]["queryResults"][0] is None




from enum import Enum
from typing import Union, Any, Optional
from dataclasses import dataclass, asdict

# Constants
IS_EXACTLY = 'is exactly'
IS_NOT_EXACTLY = 'is not exactly'
CONTAINS = 'contains'
IS = 'is'
IS_NOT = 'is not'
IS_NOT_NULL = 'is not null'
IS_NULL = 'is null'

IN_THE_LAST = 'in the last'
IN_THE_PREVIOUS = 'in the previous'
IN_THE_CURRENT = 'in the current'

EQUAL_TO = 'equal to'
NOT_EQUAL_TO = 'not equal to'
GREATER_THAN = 'greater than'
LESS_THAN = 'less than'
GREATER_THAN_OR_EQUAL_TO = 'greater than or equal to'
LESS_THAN_OR_EQUAL_TO = 'less than or equal to'

YEAR = 'year'
QUARTER = 'quarter'
MONTH = 'month'
WEEK = 'week'
DAY = 'day'
HOUR = 'hour'

NUMBER = 'number'
STRING = 'string'
DATE = 'date'
NULL = 'null'
CUSTOM = 'custom'
BOOLEAN = 'boolean'

# Enums
class StringOperator(Enum):
    IS_EXACTLY = IS_EXACTLY
    IS_NOT_EXACTLY = IS_NOT_EXACTLY
    CONTAINS = CONTAINS
    IS = IS
    IS_NOT = IS_NOT

class DateOperator(Enum):
    CUSTOM = CUSTOM
    IN_THE_LAST = IN_THE_LAST
    IN_THE_PREVIOUS = IN_THE_PREVIOUS
    IN_THE_CURRENT = IN_THE_CURRENT
    EQUAL_TO = EQUAL_TO
    NOT_EQUAL_TO = NOT_EQUAL_TO
    GREATER_THAN = GREATER_THAN
    LESS_THAN = LESS_THAN
    GREATER_THAN_OR_EQUAL_TO = GREATER_THAN_OR_EQUAL_TO
    LESS_THAN_OR_EQUAL_TO = LESS_THAN_OR_EQUAL_TO

class NumberOperator(Enum):
    EQUAL_TO = EQUAL_TO
    NOT_EQUAL_TO = NOT_EQUAL_TO
    GREATER_THAN = GREATER_THAN
    LESS_THAN = LESS_THAN
    GREATER_THAN_OR_EQUAL_TO = GREATER_THAN_OR_EQUAL_TO
    LESS_THAN_OR_EQUAL_TO = LESS_THAN_OR_EQUAL_TO

class NullOperator(Enum):
    IS_NOT_NULL = IS_NOT_NULL
    IS_NULL = IS_NULL

class BoolOperator(Enum):
    EQUAL_TO = EQUAL_TO
    NOT_EQUAL_TO = NOT_EQUAL_TO

class TimeUnit(Enum):
    YEAR = YEAR
    QUARTER = QUARTER
    MONTH = MONTH
    WEEK = WEEK
    DAY = DAY
    HOUR = HOUR

class FieldType(Enum):
    STRING = STRING
    NUMBER = NUMBER
    DATE = DATE
    NULL = NULL
    BOOLEAN = BOOLEAN

class FilterType(Enum):
    STRING_FILTER = 'string-filter'
    DATE_FILTER = 'date-filter'
    DATE_CUSTOM_FILTER = 'date-custom-filter'
    DATE_COMPARISON_FILTER = 'date-comparison-filter'
    NUMERIC_FILTER = 'numeric-filter'
    NULL_FILTER = 'null-filter'
    STRING_IN_FILTER = 'string-in-filter'
    BOOLEAN_FILTER = 'boolean-filter'

# Types
Operator = Union[StringOperator, DateOperator, NumberOperator, NullOperator, BoolOperator]

# Base Filter Interface
@dataclass
class DateRange:
    startDate: str
    endDate: str

@dataclass
class DateValue:
    value: int
    unit: TimeUnit
@dataclass
class BaseFilter:
    filterType: FilterType
    fieldType: FieldType
    operator: Operator
    field: str
    value: Union[bool, int, str, list[str], DateRange, DateValue, None]
    table: Optional[str] = None

@dataclass
class Filter:
    filter_type: FilterType
    operator: Operator
    value: Union[bool, int, str, list[str], DateRange, DateValue, None]
    field: str
    table: str

def convert_custom_filter(filter: Filter) -> dict:
    if filter.filter_type == FilterType.STRING_FILTER:
        if not isinstance(filter.value, str):
            raise ValueError('Invalid value for StringFilter, expected string')
        if filter.operator not in StringOperator:
            raise ValueError('Invalid operator for StringFilter, expected StringOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.STRING, filter.operator, filter.field, filter.value, filter.table)) 
    elif filter.filter_type == FilterType.STRING_IN_FILTER:
        if not isinstance(filter.value, list):
            raise ValueError('Invalid value for StringInFilter, expected list')
        if filter.operator not in StringOperator:
            raise ValueError('Invalid operator for StringInFilter, expected StringOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.STRING, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.NUMERIC_FILTER:
        if not isinstance(filter.value, int):
            raise ValueError('Invalid value for NumericFilter, expected int')
        if filter.operator not in NumberOperator:
            raise ValueError('Invalid operator for NumericFilter, expected NumberOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.NUMBER, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.DATE_FILTER:
        if not isinstance(filter.value, DateValue) or filter.value is None:
            raise ValueError('Invalid value for DateFilter, expected DateValue')
        if filter.operator not in DateOperator:
            raise ValueError('Invalid operator for DateFilter, expected DateOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.DATE, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.DATE_CUSTOM_FILTER:
        if not isinstance(filter.value, DateRange) or filter.value is None:
            raise ValueError('Invalid value for DateCustomFilter, expected DateRange')
        if filter.operator not in DateOperator:
            raise ValueError('Invalid operator for DateCustomFilter, expected DateOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.DATE, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.DATE_COMPARISON_FILTER:
        if not isinstance(filter.value, str):
            raise ValueError('Invalid value for DateComparisonFilter, expected str')
        if filter.operator not in DateOperator:
            raise ValueError('Invalid operator for DateComparisonFilter, expected DateOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.DATE, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.NULL_FILTER:
        if filter.value is not None:
            raise ValueError('Invalid value for NullFilter, expected None')
        if filter.operator not in NullOperator:
            raise ValueError('Invalid operator for NullFilter, expected NullOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.NULL, filter.operator, filter.field, filter.value, filter.table))
    elif filter.filter_type == FilterType.BOOLEAN_FILTER:
        if not isinstance(filter.value, bool):
            raise ValueError('Invalid value for BooleanFilter, expected bool')
        if filter.operator not in BoolOperator:
            raise ValueError('Invalid operator for BooleanFilter, expected BoolOperator')
        return asdict(BaseFilter(filter.filter_type, FieldType.BOOLEAN, filter.operator, filter.field, filter.value, filter.table))
    else:
        raise ValueError(f'Unknown filter type: {filter.filter_type}')


PG_TYPES = [
    {"typname": "bool", "oid": 16},
    {"typname": "bytea", "oid": 17},
    {"typname": "char", "oid": 18},
    {"typname": "name", "oid": 19},
    {"typname": "int8", "oid": 20},
    {"typname": "int2", "oid": 21},
    {"typname": "int2vector", "oid": 22},
    {"typname": "int4", "oid": 23},
    {"typname": "regproc", "oid": 24},
    {"typname": "text", "oid": 25},
    {"typname": "oid", "oid": 26},
    {"typname": "tid", "oid": 27},
    {"typname": "xid", "oid": 28},
    {"typname": "cid", "oid": 29},
    {"typname": "oidvector", "oid": 30},
    {"typname": "pg_ddl_command", "oid": 32},
    {"typname": "pg_type", "oid": 71},
    {"typname": "pg_attribute", "oid": 75},
    {"typname": "pg_proc", "oid": 81},
    {"typname": "pg_class", "oid": 83},
    {"typname": "json", "oid": 114},
    {"typname": "xml", "oid": 142},
    {"typname": "_xml", "oid": 143},
    {"typname": "pg_node_tree", "oid": 194},
    {"typname": "_json", "oid": 199},
    {"typname": "_pg_type", "oid": 210},
    {"typname": "table_am_handler", "oid": 269},
    {"typname": "_pg_attribute", "oid": 270},
    {"typname": "_xid8", "oid": 271},
    {"typname": "_pg_proc", "oid": 272},
    {"typname": "_pg_class", "oid": 273},
    {"typname": "index_am_handler", "oid": 325},
    {"typname": "point", "oid": 600},
    {"typname": "lseg", "oid": 601},
    {"typname": "path", "oid": 602},
    {"typname": "box", "oid": 603},
    {"typname": "polygon", "oid": 604},
    {"typname": "line", "oid": 628},
    {"typname": "_line", "oid": 629},
    {"typname": "cidr", "oid": 650},
    {"typname": "_cidr", "oid": 651},
    {"typname": "float4", "oid": 700},
    {"typname": "float8", "oid": 701},
    {"typname": "unknown", "oid": 705},
    {"typname": "circle", "oid": 718},
    {"typname": "_circle", "oid": 719},
    {"typname": "macaddr8", "oid": 774},
    {"typname": "_macaddr8", "oid": 775},
    {"typname": "money", "oid": 790},
    {"typname": "_money", "oid": 791},
    {"typname": "macaddr", "oid": 829},
    {"typname": "inet", "oid": 869},
    {"typname": "_bool", "oid": 1000},
    {"typname": "_bytea", "oid": 1001},
    {"typname": "_char", "oid": 1002},
    {"typname": "_name", "oid": 1003},
    {"typname": "_int2", "oid": 1005},
    {"typname": "_int2vector", "oid": 1006},
    {"typname": "_int4", "oid": 1007},
    {"typname": "_regproc", "oid": 1008},
    {"typname": "_text", "oid": 1009},
    {"typname": "_tid", "oid": 1010},
    {"typname": "_xid", "oid": 1011},
    {"typname": "_cid", "oid": 1012},
    {"typname": "_oidvector", "oid": 1013},
    {"typname": "_bpchar", "oid": 1014},
    {"typname": "_varchar", "oid": 1015},
    {"typname": "_int8", "oid": 1016},
    {"typname": "_point", "oid": 1017},
    {"typname": "_lseg", "oid": 1018},
    {"typname": "_path", "oid": 1019},
    {"typname": "_box", "oid": 1020},
    {"typname": "_float4", "oid": 1021},
    {"typname": "_float8", "oid": 1022},
    {"typname": "_polygon", "oid": 1027},
    {"typname": "_oid", "oid": 1028},
    {"typname": "aclitem", "oid": 1033},
    {"typname": "_aclitem", "oid": 1034},
    {"typname": "_macaddr", "oid": 1040},
    {"typname": "_inet", "oid": 1041},
    {"typname": "bpchar", "oid": 1042},
    {"typname": "varchar", "oid": 1043},
    {"typname": "date", "oid": 1082},
    {"typname": "time", "oid": 1083},
    {"typname": "timestamp", "oid": 1114},
    {"typname": "_timestamp", "oid": 1115},
    {"typname": "_date", "oid": 1182},
    {"typname": "_time", "oid": 1183},
    {"typname": "timestamptz", "oid": 1184},
    {"typname": "_timestamptz", "oid": 1185},
    {"typname": "interval", "oid": 1186},
    {"typname": "_interval", "oid": 1187},
    {"typname": "_numeric", "oid": 1231},
    {"typname": "pg_database", "oid": 1248},
    {"typname": "_cstring", "oid": 1263},
    {"typname": "timetz", "oid": 1266},
    {"typname": "_timetz", "oid": 1270},
    {"typname": "bit", "oid": 1560},
    {"typname": "_bit", "oid": 1561},
    {"typname": "varbit", "oid": 1562},
    {"typname": "_varbit", "oid": 1563},
    {"typname": "numeric", "oid": 1700},
    {"typname": "refcursor", "oid": 1790},
    {"typname": "_refcursor", "oid": 2201},
    {"typname": "regprocedure", "oid": 2202},
    {"typname": "regoper", "oid": 2203},
    {"typname": "regoperator", "oid": 2204},
    {"typname": "regclass", "oid": 2205},
    {"typname": "regtype", "oid": 2206},
    {"typname": "_regprocedure", "oid": 2207},
    {"typname": "_regoper", "oid": 2208},
    {"typname": "_regoperator", "oid": 2209},
    {"typname": "_regclass", "oid": 2210},
    {"typname": "_regtype", "oid": 2211},
    {"typname": "record", "oid": 2249},
    {"typname": "cstring", "oid": 2275},
    {"typname": "any", "oid": 2276},
    {"typname": "anyarray", "oid": 2277},
    {"typname": "void", "oid": 2278},
    {"typname": "trigger", "oid": 2279},
    {"typname": "language_handler", "oid": 2280},
    {"typname": "internal", "oid": 2281},
    {"typname": "anyelement", "oid": 2283},
    {"typname": "_record", "oid": 2287},
    {"typname": "anynonarray", "oid": 2776},
    {"typname": "pg_authid", "oid": 2842},
    {"typname": "pg_auth_members", "oid": 2843},
    {"typname": "_txid_snapshot", "oid": 2949},
    {"typname": "uuid", "oid": 2950},
    {"typname": "_uuid", "oid": 2951},
    {"typname": "txid_snapshot", "oid": 2970},
    {"typname": "fdw_handler", "oid": 3115},
    {"typname": "pg_lsn", "oid": 3220},
    {"typname": "_pg_lsn", "oid": 3221},
    {"typname": "tsm_handler", "oid": 3310},
    {"typname": "pg_ndistinct", "oid": 3361},
    {"typname": "pg_dependencies", "oid": 3402},
    {"typname": "anyenum", "oid": 3500},
    {"typname": "tsvector", "oid": 3614},
    {"typname": "tsquery", "oid": 3615},
    {"typname": "gtsvector", "oid": 3642},
    {"typname": "_tsvector", "oid": 3643},
    {"typname": "_gtsvector", "oid": 3644},
    {"typname": "_tsquery", "oid": 3645},
    {"typname": "regconfig", "oid": 3734},
    {"typname": "_regconfig", "oid": 3735},
    {"typname": "regdictionary", "oid": 3769},
    {"typname": "_regdictionary", "oid": 3770},
    {"typname": "jsonb", "oid": 3802},
    {"typname": "_jsonb", "oid": 3807},
    {"typname": "anyrange", "oid": 3831},
    {"typname": "event_trigger", "oid": 3838},
    {"typname": "int4range", "oid": 3904},
    {"typname": "_int4range", "oid": 3905},
    {"typname": "numrange", "oid": 3906},
    {"typname": "_numrange", "oid": 3907},
    {"typname": "tsrange", "oid": 3908},
    {"typname": "_tsrange", "oid": 3909},
    {"typname": "tstzrange", "oid": 3910},
    {"typname": "_tstzrange", "oid": 3911},
    {"typname": "daterange", "oid": 3912},
    {"typname": "_daterange", "oid": 3913},
    {"typname": "int8range", "oid": 3926},
    {"typname": "_int8range", "oid": 3927},
    {"typname": "pg_shseclabel", "oid": 4066},
    {"typname": "jsonpath", "oid": 4072},
    {"typname": "_jsonpath", "oid": 4073},
    {"typname": "regnamespace", "oid": 4089},
    {"typname": "_regnamespace", "oid": 4090},
    {"typname": "regrole", "oid": 4096},
    {"typname": "_regrole", "oid": 4097},
    {"typname": "regcollation", "oid": 4191},
    {"typname": "_regcollation", "oid": 4192},
    {"typname": "int4multirange", "oid": 4451},
    {"typname": "nummultirange", "oid": 4532},
    {"typname": "tsmultirange", "oid": 4533},
    {"typname": "tstzmultirange", "oid": 4534},
    {"typname": "datemultirange", "oid": 4535},
    {"typname": "int8multirange", "oid": 4536},
    {"typname": "anymultirange", "oid": 4537},
    {"typname": "anycompatiblemultirange", "oid": 4538},
    {"typname": "pg_brin_bloom_summary", "oid": 4600},
    {"typname": "pg_brin_minmax_multi_summary", "oid": 4601},
    {"typname": "pg_mcv_list", "oid": 5017},
    {"typname": "pg_snapshot", "oid": 5038},
    {"typname": "_pg_snapshot", "oid": 5039},
    {"typname": "xid8", "oid": 5069},
    {"typname": "anycompatible", "oid": 5077},
    {"typname": "anycompatiblearray", "oid": 5078},
    {"typname": "anycompatiblenonarray", "oid": 5079},
    {"typname": "anycompatiblerange", "oid": 5080},
    {"typname": "pg_subscription", "oid": 6101},
    {"typname": "_int4multirange", "oid": 6150},
    {"typname": "_nummultirange", "oid": 6151},
    {"typname": "_tsmultirange", "oid": 6152},
    {"typname": "_tstzmultirange", "oid": 6153},
    {"typname": "_datemultirange", "oid": 6155},
    {"typname": "_int8multirange", "oid": 6157},
    {"typname": "_pg_attrdef", "oid": 10000},
    {"typname": "pg_attrdef", "oid": 10001},
    {"typname": "_pg_constraint", "oid": 10002},
    {"typname": "pg_constraint", "oid": 10003},
    {"typname": "_pg_inherits", "oid": 10004},
    {"typname": "pg_inherits", "oid": 10005},
    {"typname": "_pg_index", "oid": 10006},
    {"typname": "pg_index", "oid": 10007},
    {"typname": "_pg_operator", "oid": 10008},
    {"typname": "pg_operator", "oid": 10009},
    {"typname": "_pg_opfamily", "oid": 10010},
    {"typname": "pg_opfamily", "oid": 10011},
    {"typname": "_pg_opclass", "oid": 10012},
    {"typname": "pg_opclass", "oid": 10013},
    {"typname": "_pg_am", "oid": 10014},
    {"typname": "pg_am", "oid": 10015},
    {"typname": "_pg_amop", "oid": 10016},
    {"typname": "pg_amop", "oid": 10017},
    {"typname": "_pg_amproc", "oid": 10018},
    {"typname": "pg_amproc", "oid": 10019},
    {"typname": "_pg_language", "oid": 10020},
    {"typname": "pg_language", "oid": 10021},
    {"typname": "_pg_largeobject_metadata", "oid": 10022},
    {"typname": "pg_largeobject_metadata", "oid": 10023},
    {"typname": "_pg_largeobject", "oid": 10024},
    {"typname": "pg_largeobject", "oid": 10025},
    {"typname": "_pg_aggregate", "oid": 10026},
    {"typname": "pg_aggregate", "oid": 10027},
    {"typname": "_pg_statistic", "oid": 10028},
    {"typname": "pg_statistic", "oid": 10029},
    {"typname": "_pg_statistic_ext", "oid": 10030},
    {"typname": "pg_statistic_ext", "oid": 10031},
    {"typname": "_pg_statistic_ext_data", "oid": 10032},
    {"typname": "pg_statistic_ext_data", "oid": 10033},
    {"typname": "_pg_rewrite", "oid": 10034},
    {"typname": "pg_rewrite", "oid": 10035},
    {"typname": "_pg_trigger", "oid": 10036},
    {"typname": "pg_trigger", "oid": 10037},
    {"typname": "_pg_event_trigger", "oid": 10038},
    {"typname": "pg_event_trigger", "oid": 10039},
    {"typname": "_pg_description", "oid": 10040},
    {"typname": "pg_description", "oid": 10041},
    {"typname": "_pg_cast", "oid": 10042},
    {"typname": "pg_cast", "oid": 10043},
    {"typname": "_pg_enum", "oid": 10044},
    {"typname": "pg_enum", "oid": 10045},
    {"typname": "_pg_namespace", "oid": 10046},
    {"typname": "pg_namespace", "oid": 10047},
    {"typname": "_pg_conversion", "oid": 10048},
    {"typname": "pg_conversion", "oid": 10049},
    {"typname": "_pg_depend", "oid": 10050},
    {"typname": "pg_depend", "oid": 10051},
    {"typname": "_pg_database", "oid": 10052},
    {"typname": "_pg_db_role_setting", "oid": 10053},
    {"typname": "pg_db_role_setting", "oid": 10054},
    {"typname": "_pg_tablespace", "oid": 10055},
    {"typname": "pg_tablespace", "oid": 10056},
    {"typname": "_pg_authid", "oid": 10057},
    {"typname": "_pg_auth_members", "oid": 10058},
    {"typname": "_pg_shdepend", "oid": 10059},
    {"typname": "pg_shdepend", "oid": 10060},
    {"typname": "_pg_shdescription", "oid": 10061},
    {"typname": "pg_shdescription", "oid": 10062},
    {"typname": "_pg_ts_config", "oid": 10063},
    {"typname": "pg_ts_config", "oid": 10064},
    {"typname": "_pg_ts_config_map", "oid": 10065},
    {"typname": "pg_ts_config_map", "oid": 10066},
    {"typname": "_pg_ts_dict", "oid": 10067},
    {"typname": "pg_ts_dict", "oid": 10068},
    {"typname": "_pg_ts_parser", "oid": 10069},
    {"typname": "pg_ts_parser", "oid": 10070},
    {"typname": "_pg_ts_template", "oid": 10071},
    {"typname": "pg_ts_template", "oid": 10072},
    {"typname": "_pg_extension", "oid": 10073},
    {"typname": "pg_extension", "oid": 10074},
    {"typname": "_pg_foreign_data_wrapper", "oid": 10075},
    {"typname": "pg_foreign_data_wrapper", "oid": 10076},
    {"typname": "_pg_foreign_server", "oid": 10077},
    {"typname": "pg_foreign_server", "oid": 10078},
    {"typname": "_pg_user_mapping", "oid": 10079},
    {"typname": "pg_user_mapping", "oid": 10080},
    {"typname": "_pg_foreign_table", "oid": 10081},
    {"typname": "pg_foreign_table", "oid": 10082},
    {"typname": "_pg_policy", "oid": 10083},
    {"typname": "pg_policy", "oid": 10084},
    {"typname": "_pg_replication_origin", "oid": 10085},
    {"typname": "pg_replication_origin", "oid": 10086},
    {"typname": "_pg_default_acl", "oid": 10087},
    {"typname": "pg_default_acl", "oid": 10088},
    {"typname": "_pg_init_privs", "oid": 10089},
    {"typname": "pg_init_privs", "oid": 10090},
    {"typname": "_pg_seclabel", "oid": 10091},
    {"typname": "pg_seclabel", "oid": 10092},
    {"typname": "_pg_shseclabel", "oid": 10093},
    {"typname": "_pg_collation", "oid": 10094},
    {"typname": "pg_collation", "oid": 10095},
    {"typname": "_pg_parameter_acl", "oid": 10096},
    {"typname": "pg_parameter_acl", "oid": 10097},
    {"typname": "_pg_partitioned_table", "oid": 10098},
    {"typname": "pg_partitioned_table", "oid": 10099},
    {"typname": "_pg_range", "oid": 10100},
    {"typname": "pg_range", "oid": 10101},
    {"typname": "_pg_transform", "oid": 10102},
    {"typname": "pg_transform", "oid": 10103},
    {"typname": "_pg_sequence", "oid": 10104},
    {"typname": "pg_sequence", "oid": 10105},
    {"typname": "_pg_publication", "oid": 10106},
    {"typname": "pg_publication", "oid": 10107},
    {"typname": "_pg_publication_namespace", "oid": 10108},
    {"typname": "pg_publication_namespace", "oid": 10109},
    {"typname": "_pg_publication_rel", "oid": 10110},
    {"typname": "pg_publication_rel", "oid": 10111},
    {"typname": "_pg_subscription", "oid": 10112},
    {"typname": "_pg_subscription_rel", "oid": 10113},
    {"typname": "pg_subscription_rel", "oid": 10114},
    {"typname": "_pg_roles", "oid": 12001},
    {"typname": "pg_roles", "oid": 12002},
    {"typname": "_pg_shadow", "oid": 12006},
    {"typname": "pg_shadow", "oid": 12007},
    {"typname": "_pg_group", "oid": 12011},
    {"typname": "pg_group", "oid": 12012},
    {"typname": "_pg_user", "oid": 12015},
    {"typname": "pg_user", "oid": 12016},
    {"typname": "_pg_policies", "oid": 12019},
    {"typname": "pg_policies", "oid": 12020},
    {"typname": "_pg_rules", "oid": 12024},
    {"typname": "pg_rules", "oid": 12025},
    {"typname": "_pg_views", "oid": 12029},
    {"typname": "pg_views", "oid": 12030},
    {"typname": "_pg_tables", "oid": 12034},
    {"typname": "pg_tables", "oid": 12035},
    {"typname": "_pg_matviews", "oid": 12039},
    {"typname": "pg_matviews", "oid": 12040},
    {"typname": "_pg_indexes", "oid": 12044},
    {"typname": "pg_indexes", "oid": 12045},
    {"typname": "_pg_sequences", "oid": 12049},
    {"typname": "pg_sequences", "oid": 12050},
    {"typname": "_pg_stats", "oid": 12054},
    {"typname": "pg_stats", "oid": 12055},
    {"typname": "_pg_stats_ext", "oid": 12059},
    {"typname": "pg_stats_ext", "oid": 12060},
    {"typname": "_pg_stats_ext_exprs", "oid": 12064},
    {"typname": "pg_stats_ext_exprs", "oid": 12065},
    {"typname": "_pg_publication_tables", "oid": 12069},
    {"typname": "pg_publication_tables", "oid": 12070},
    {"typname": "_pg_locks", "oid": 12074},
    {"typname": "pg_locks", "oid": 12075},
    {"typname": "_pg_cursors", "oid": 12078},
    {"typname": "pg_cursors", "oid": 12079},
    {"typname": "_pg_available_extensions", "oid": 12082},
    {"typname": "pg_available_extensions", "oid": 12083},
    {"typname": "_pg_available_extension_versions", "oid": 12086},
    {"typname": "pg_available_extension_versions", "oid": 12087},
    {"typname": "_pg_prepared_xacts", "oid": 12091},
    {"typname": "pg_prepared_xacts", "oid": 12092},
    {"typname": "_pg_prepared_statements", "oid": 12096},
    {"typname": "pg_prepared_statements", "oid": 12097},
    {"typname": "_pg_seclabels", "oid": 12100},
    {"typname": "pg_seclabels", "oid": 12101},
    {"typname": "_pg_settings", "oid": 12105},
    {"typname": "pg_settings", "oid": 12106},
    {"typname": "_pg_file_settings", "oid": 12111},
    {"typname": "pg_file_settings", "oid": 12112},
    {"typname": "_pg_hba_file_rules", "oid": 12115},
    {"typname": "pg_hba_file_rules", "oid": 12116},
    {"typname": "_pg_ident_file_mappings", "oid": 12119},
    {"typname": "pg_ident_file_mappings", "oid": 12120},
    {"typname": "_pg_timezone_abbrevs", "oid": 12123},
    {"typname": "pg_timezone_abbrevs", "oid": 12124},
    {"typname": "_pg_timezone_names", "oid": 12127},
    {"typname": "pg_timezone_names", "oid": 12128},
    {"typname": "_pg_config", "oid": 12131},
    {"typname": "pg_config", "oid": 12132},
    {"typname": "_pg_shmem_allocations", "oid": 12135},
    {"typname": "pg_shmem_allocations", "oid": 12136},
    {"typname": "_pg_backend_memory_contexts", "oid": 12139},
    {"typname": "pg_backend_memory_contexts", "oid": 12140},
    {"typname": "_pg_stat_all_tables", "oid": 12143},
    {"typname": "pg_stat_all_tables", "oid": 12144},
    {"typname": "_pg_stat_xact_all_tables", "oid": 12148},
    {"typname": "pg_stat_xact_all_tables", "oid": 12149},
    {"typname": "_pg_stat_sys_tables", "oid": 12153},
    {"typname": "pg_stat_sys_tables", "oid": 12154},
    {"typname": "_pg_stat_xact_sys_tables", "oid": 12158},
    {"typname": "pg_stat_xact_sys_tables", "oid": 12159},
    {"typname": "_pg_stat_user_tables", "oid": 12162},
    {"typname": "pg_stat_user_tables", "oid": 12163},
    {"typname": "_pg_stat_xact_user_tables", "oid": 12167},
    {"typname": "pg_stat_xact_user_tables", "oid": 12168},
    {"typname": "_pg_statio_all_tables", "oid": 12171},
    {"typname": "pg_statio_all_tables", "oid": 12172},
    {"typname": "_pg_statio_sys_tables", "oid": 12176},
    {"typname": "pg_statio_sys_tables", "oid": 12177},
    {"typname": "_pg_statio_user_tables", "oid": 12180},
    {"typname": "pg_statio_user_tables", "oid": 12181},
    {"typname": "_pg_stat_all_indexes", "oid": 12184},
    {"typname": "pg_stat_all_indexes", "oid": 12185},
    {"typname": "_pg_stat_sys_indexes", "oid": 12189},
    {"typname": "pg_stat_sys_indexes", "oid": 12190},
    {"typname": "_pg_stat_user_indexes", "oid": 12193},
    {"typname": "pg_stat_user_indexes", "oid": 12194},
    {"typname": "_pg_statio_all_indexes", "oid": 12197},
    {"typname": "pg_statio_all_indexes", "oid": 12198},
    {"typname": "_pg_statio_sys_indexes", "oid": 12202},
    {"typname": "pg_statio_sys_indexes", "oid": 12203},
    {"typname": "_pg_statio_user_indexes", "oid": 12206},
    {"typname": "pg_statio_user_indexes", "oid": 12207},
    {"typname": "_pg_statio_all_sequences", "oid": 12210},
    {"typname": "pg_statio_all_sequences", "oid": 12211},
    {"typname": "_pg_statio_sys_sequences", "oid": 12215},
    {"typname": "pg_statio_sys_sequences", "oid": 12216},
    {"typname": "_pg_statio_user_sequences", "oid": 12219},
    {"typname": "pg_statio_user_sequences", "oid": 12220},
    {"typname": "_pg_stat_activity", "oid": 12223},
    {"typname": "pg_stat_activity", "oid": 12224},
    {"typname": "_pg_stat_replication", "oid": 12228},
    {"typname": "pg_stat_replication", "oid": 12229},
    {"typname": "_pg_stat_slru", "oid": 12233},
    {"typname": "pg_stat_slru", "oid": 12234},
    {"typname": "_pg_stat_wal_receiver", "oid": 12237},
    {"typname": "pg_stat_wal_receiver", "oid": 12238},
    {"typname": "_pg_stat_recovery_prefetch", "oid": 12241},
    {"typname": "pg_stat_recovery_prefetch", "oid": 12242},
    {"typname": "_pg_stat_subscription", "oid": 12245},
    {"typname": "pg_stat_subscription", "oid": 12246},
    {"typname": "_pg_stat_ssl", "oid": 12250},
    {"typname": "pg_stat_ssl", "oid": 12251},
    {"typname": "_pg_stat_gssapi", "oid": 12254},
    {"typname": "pg_stat_gssapi", "oid": 12255},
    {"typname": "_pg_replication_slots", "oid": 12258},
    {"typname": "pg_replication_slots", "oid": 12259},
    {"typname": "_pg_stat_replication_slots", "oid": 12263},
    {"typname": "pg_stat_replication_slots", "oid": 12264},
    {"typname": "_pg_stat_database", "oid": 12267},
    {"typname": "pg_stat_database", "oid": 12268},
    {"typname": "_pg_stat_database_conflicts", "oid": 12272},
    {"typname": "pg_stat_database_conflicts", "oid": 12273},
    {"typname": "_pg_stat_user_functions", "oid": 12276},
    {"typname": "pg_stat_user_functions", "oid": 12277},
    {"typname": "_pg_stat_xact_user_functions", "oid": 12281},
    {"typname": "pg_stat_xact_user_functions", "oid": 12282},
    {"typname": "_pg_stat_archiver", "oid": 12286},
    {"typname": "pg_stat_archiver", "oid": 12287},
    {"typname": "_pg_stat_bgwriter", "oid": 12290},
    {"typname": "pg_stat_bgwriter", "oid": 12291},
    {"typname": "_pg_stat_wal", "oid": 12294},
    {"typname": "pg_stat_wal", "oid": 12295},
    {"typname": "_pg_stat_progress_analyze", "oid": 12298},
    {"typname": "pg_stat_progress_analyze", "oid": 12299},
    {"typname": "_pg_stat_progress_vacuum", "oid": 12303},
    {"typname": "pg_stat_progress_vacuum", "oid": 12304},
    {"typname": "_pg_stat_progress_cluster", "oid": 12308},
    {"typname": "pg_stat_progress_cluster", "oid": 12309},
    {"typname": "_pg_stat_progress_create_index", "oid": 12313},
    {"typname": "pg_stat_progress_create_index", "oid": 12314},
    {"typname": "_pg_stat_progress_basebackup", "oid": 12318},
    {"typname": "pg_stat_progress_basebackup", "oid": 12319},
    {"typname": "_pg_stat_progress_copy", "oid": 12323},
    {"typname": "pg_stat_progress_copy", "oid": 12324},
    {"typname": "_pg_user_mappings", "oid": 12328},
    {"typname": "pg_user_mappings", "oid": 12329},
    {"typname": "_pg_replication_origin_status", "oid": 12333},
    {"typname": "pg_replication_origin_status", "oid": 12334},
    {"typname": "_pg_stat_subscription_stats", "oid": 12337},
    {"typname": "pg_stat_subscription_stats", "oid": 12338},
    {"typname": "_cardinal_number", "oid": 12416},
    {"typname": "cardinal_number", "oid": 12417},
    {"typname": "_character_data", "oid": 12419},
    {"typname": "character_data", "oid": 12420},
    {"typname": "_sql_identifier", "oid": 12421},
    {"typname": "sql_identifier", "oid": 12422},
    {"typname": "_information_schema_catalog_name", "oid": 12424},
    {"typname": "information_schema_catalog_name", "oid": 12425},
    {"typname": "_time_stamp", "oid": 12427},
    {"typname": "time_stamp", "oid": 12428},
    {"typname": "_yes_or_no", "oid": 12429},
    {"typname": "yes_or_no", "oid": 12430},
    {"typname": "_applicable_roles", "oid": 12433},
    {"typname": "applicable_roles", "oid": 12434},
    {"typname": "_administrable_role_authorizations", "oid": 12438},
    {"typname": "administrable_role_authorizations", "oid": 12439},
    {"typname": "_attributes", "oid": 12442},
    {"typname": "attributes", "oid": 12443},
    {"typname": "_character_sets", "oid": 12447},
    {"typname": "character_sets", "oid": 12448},
    {"typname": "_check_constraint_routine_usage", "oid": 12452},
    {"typname": "check_constraint_routine_usage", "oid": 12453},
    {"typname": "_check_constraints", "oid": 12457},
    {"typname": "check_constraints", "oid": 12458},
    {"typname": "_collations", "oid": 12462},
    {"typname": "collations", "oid": 12463},
    {"typname": "_collation_character_set_applicability", "oid": 12467},
    {"typname": "collation_character_set_applicability", "oid": 12468},
    {"typname": "_column_column_usage", "oid": 12472},
    {"typname": "column_column_usage", "oid": 12473},
    {"typname": "_column_domain_usage", "oid": 12477},
    {"typname": "column_domain_usage", "oid": 12478},
    {"typname": "_column_privileges", "oid": 12482},
    {"typname": "column_privileges", "oid": 12483},
    {"typname": "_column_udt_usage", "oid": 12487},
    {"typname": "column_udt_usage", "oid": 12488},
    {"typname": "_columns", "oid": 12492},
    {"typname": "columns", "oid": 12493},
    {"typname": "_constraint_column_usage", "oid": 12497},
    {"typname": "constraint_column_usage", "oid": 12498},
    {"typname": "_constraint_table_usage", "oid": 12502},
    {"typname": "constraint_table_usage", "oid": 12503},
    {"typname": "_domain_constraints", "oid": 12507},
    {"typname": "domain_constraints", "oid": 12508},
    {"typname": "_domain_udt_usage", "oid": 12512},
    {"typname": "domain_udt_usage", "oid": 12513},
    {"typname": "_domains", "oid": 12517},
    {"typname": "domains", "oid": 12518},
    {"typname": "_enabled_roles", "oid": 12522},
    {"typname": "enabled_roles", "oid": 12523},
    {"typname": "_key_column_usage", "oid": 12526},
    {"typname": "key_column_usage", "oid": 12527},
    {"typname": "_parameters", "oid": 12531},
    {"typname": "parameters", "oid": 12532},
    {"typname": "_referential_constraints", "oid": 12536},
    {"typname": "referential_constraints", "oid": 12537},
    {"typname": "_role_column_grants", "oid": 12541},
    {"typname": "role_column_grants", "oid": 12542},
    {"typname": "_routine_column_usage", "oid": 12545},
    {"typname": "routine_column_usage", "oid": 12546},
    {"typname": "_routine_privileges", "oid": 12550},
    {"typname": "routine_privileges", "oid": 12551},
    {"typname": "_role_routine_grants", "oid": 12555},
    {"typname": "role_routine_grants", "oid": 12556},
    {"typname": "_routine_routine_usage", "oid": 12559},
    {"typname": "routine_routine_usage", "oid": 12560},
    {"typname": "_routine_sequence_usage", "oid": 12564},
    {"typname": "routine_sequence_usage", "oid": 12565},
    {"typname": "_routine_table_usage", "oid": 12569},
    {"typname": "routine_table_usage", "oid": 12570},
    {"typname": "_routines", "oid": 12574},
    {"typname": "routines", "oid": 12575},
    {"typname": "_schemata", "oid": 12579},
    {"typname": "schemata", "oid": 12580},
    {"typname": "_sequences", "oid": 12583},
    {"typname": "sequences", "oid": 12584},
    {"typname": "_sql_features", "oid": 12588},
    {"typname": "sql_features", "oid": 12589},
    {"typname": "_sql_implementation_info", "oid": 12593},
    {"typname": "sql_implementation_info", "oid": 12594},
    {"typname": "_sql_parts", "oid": 12598},
    {"typname": "sql_parts", "oid": 12599},
    {"typname": "_sql_sizing", "oid": 12603},
    {"typname": "sql_sizing", "oid": 12604},
    {"typname": "_table_constraints", "oid": 12608},
    {"typname": "table_constraints", "oid": 12609},
    {"typname": "_table_privileges", "oid": 12613},
    {"typname": "table_privileges", "oid": 12614},
    {"typname": "_role_table_grants", "oid": 12618},
    {"typname": "role_table_grants", "oid": 12619},
    {"typname": "_tables", "oid": 12622},
    {"typname": "tables", "oid": 12623},
    {"typname": "_transforms", "oid": 12627},
    {"typname": "transforms", "oid": 12628},
    {"typname": "_triggered_update_columns", "oid": 12632},
    {"typname": "triggered_update_columns", "oid": 12633},
    {"typname": "_triggers", "oid": 12637},
    {"typname": "triggers", "oid": 12638},
    {"typname": "_udt_privileges", "oid": 12642},
    {"typname": "udt_privileges", "oid": 12643},
    {"typname": "_role_udt_grants", "oid": 12647},
    {"typname": "role_udt_grants", "oid": 12648},
    {"typname": "_usage_privileges", "oid": 12651},
    {"typname": "usage_privileges", "oid": 12652},
    {"typname": "_role_usage_grants", "oid": 12656},
    {"typname": "role_usage_grants", "oid": 12657},
    {"typname": "_user_defined_types", "oid": 12660},
    {"typname": "user_defined_types", "oid": 12661},
    {"typname": "_view_column_usage", "oid": 12665},
    {"typname": "view_column_usage", "oid": 12666},
    {"typname": "_view_routine_usage", "oid": 12670},
    {"typname": "view_routine_usage", "oid": 12671},
    {"typname": "_view_table_usage", "oid": 12675},
    {"typname": "view_table_usage", "oid": 12676},
    {"typname": "_views", "oid": 12680},
    {"typname": "views", "oid": 12681},
    {"typname": "_data_type_privileges", "oid": 12685},
    {"typname": "data_type_privileges", "oid": 12686},
    {"typname": "_element_types", "oid": 12690},
    {"typname": "element_types", "oid": 12691},
    {"typname": "__pg_foreign_table_columns", "oid": 12695},
    {"typname": "_pg_foreign_table_columns", "oid": 12696},
    {"typname": "_column_options", "oid": 12700},
    {"typname": "column_options", "oid": 12701},
    {"typname": "__pg_foreign_data_wrappers", "oid": 12704},
    {"typname": "_pg_foreign_data_wrappers", "oid": 12705},
    {"typname": "_foreign_data_wrapper_options", "oid": 12708},
    {"typname": "foreign_data_wrapper_options", "oid": 12709},
    {"typname": "_foreign_data_wrappers", "oid": 12712},
    {"typname": "foreign_data_wrappers", "oid": 12713},
    {"typname": "__pg_foreign_servers", "oid": 12716},
    {"typname": "_pg_foreign_servers", "oid": 12717},
    {"typname": "_foreign_server_options", "oid": 12721},
    {"typname": "foreign_server_options", "oid": 12722},
    {"typname": "_foreign_servers", "oid": 12725},
    {"typname": "foreign_servers", "oid": 12726},
    {"typname": "__pg_foreign_tables", "oid": 12729},
    {"typname": "_pg_foreign_tables", "oid": 12730},
    {"typname": "_foreign_table_options", "oid": 12734},
    {"typname": "foreign_table_options", "oid": 12735},
    {"typname": "_foreign_tables", "oid": 12738},
    {"typname": "foreign_tables", "oid": 12739},
    {"typname": "__pg_user_mappings", "oid": 12742},
    {"typname": "_pg_user_mappings", "oid": 12743},
    {"typname": "_user_mapping_options", "oid": 12747},
    {"typname": "user_mapping_options", "oid": 12748},
    {"typname": "_user_mappings", "oid": 12752},
    {"typname": "user_mappings", "oid": 12753},
    {"typname": "_crypto_box_keypair", "oid": 16657},
    {"typname": "crypto_box_keypair", "oid": 16658},
    {"typname": "_crypto_sign_keypair", "oid": 16664},
    {"typname": "crypto_sign_keypair", "oid": 16665},
    {"typname": "_crypto_kx_keypair", "oid": 16680},
    {"typname": "crypto_kx_keypair", "oid": 16681},
    {"typname": "_crypto_kx_session", "oid": 16686},
    {"typname": "crypto_kx_session", "oid": 16687},
    {"typname": "_crypto_signcrypt_state_key", "oid": 16754},
    {"typname": "crypto_signcrypt_state_key", "oid": 16755},
    {"typname": "_crypto_signcrypt_keypair", "oid": 16757},
    {"typname": "crypto_signcrypt_keypair", "oid": 16758},
    {"typname": "_key_status", "oid": 16771},
    {"typname": "key_status", "oid": 16772},
    {"typname": "_key_type", "oid": 16781},
    {"typname": "key_type", "oid": 16782},
    {"typname": "_key", "oid": 16789},
    {"typname": "key", "oid": 16790},
    {"typname": "__key_id_context", "oid": 16810},
    {"typname": "_key_id_context", "oid": 16811},
    {"typname": "_valid_key", "oid": 16893},
    {"typname": "valid_key", "oid": 16894},
    {"typname": "_masking_rule", "oid": 16909},
    {"typname": "masking_rule", "oid": 16910},
    {"typname": "_mask_columns", "oid": 16914},
    {"typname": "mask_columns", "oid": 16915},
    {"typname": "_decrypted_key", "oid": 16939},
    {"typname": "decrypted_key", "oid": 16940},
    {"typname": "_pg_stat_statements_info", "oid": 27035},
    {"typname": "pg_stat_statements_info", "oid": 27036},
    {"typname": "_pg_stat_statements", "oid": 27046},
    {"typname": "pg_stat_statements", "oid": 27047},
    {"typname": "_aal_level", "oid": 27105},
    {"typname": "aal_level", "oid": 27106},
    {"typname": "_factor_status", "oid": 27113},
    {"typname": "factor_status", "oid": 27114},
    {"typname": "_factor_type", "oid": 27119},
    {"typname": "factor_type", "oid": 27120},
    {"typname": "_audit_log_entries", "oid": 27142},
    {"typname": "audit_log_entries", "oid": 27143},
    {"typname": "_identities", "oid": 27148},
    {"typname": "identities", "oid": 27149},
    {"typname": "_instances", "oid": 27153},
    {"typname": "instances", "oid": 27154},
    {"typname": "_mfa_amr_claims", "oid": 27158},
    {"typname": "mfa_amr_claims", "oid": 27159},
    {"typname": "_mfa_challenges", "oid": 27163},
    {"typname": "mfa_challenges", "oid": 27164},
    {"typname": "_mfa_factors", "oid": 27168},
    {"typname": "mfa_factors", "oid": 27169},
    {"typname": "_refresh_tokens", "oid": 27173},
    {"typname": "refresh_tokens", "oid": 27174},
    {"typname": "_saml_providers", "oid": 27179},
    {"typname": "saml_providers", "oid": 27180},
    {"typname": "_saml_relay_states", "oid": 27187},
    {"typname": "saml_relay_states", "oid": 27188},
    {"typname": "_schema_migrations", "oid": 27193},
    {"typname": "schema_migrations", "oid": 27194},
    {"typname": "_sessions", "oid": 27196},
    {"typname": "sessions", "oid": 27197},
    {"typname": "_sso_domains", "oid": 27199},
    {"typname": "sso_domains", "oid": 27200},
    {"typname": "_sso_providers", "oid": 27205},
    {"typname": "sso_providers", "oid": 27206},
    {"typname": "_users", "oid": 27215},
    {"typname": "users", "oid": 27216},
    {"typname": "_invoice", "oid": 27228},
    {"typname": "invoice", "oid": 27229},
    {"typname": "_subscription", "oid": 27234},
    {"typname": "subscription", "oid": 27235},
    {"typname": "_transactions", "oid": 27240},
    {"typname": "transactions", "oid": 27241},
    {"typname": "_buckets", "oid": 27247},
    {"typname": "buckets", "oid": 27248},
    {"typname": "_migrations", "oid": 27255},
    {"typname": "migrations", "oid": 27256},
    {"typname": "_objects", "oid": 27259},
    {"typname": "objects", "oid": 27260},
    {"typname": "_code_challenge_method", "oid": 27516},
    {"typname": "code_challenge_method", "oid": 27517},
    {"typname": "_flow_state", "oid": 27522},
    {"typname": "flow_state", "oid": 27523},
    {"typname": "_customers", "oid": 27531},
    {"typname": "customers", "oid": 27532},
]



from google.cloud import bigquery
from google.oauth2 import service_account
import json
import re


def format_bigquery_config(connection_string):
    # find the start of the json {
    json_start = connection_string.find("{")
    if json_start == -1:
        raise Exception("Invalid input string. No JSON data found.")

    dataset_name = connection_string[0:json_start].strip()
    json_string = connection_string[json_start:]
    try:
        service_account = json.loads(json_string)
        if not service_account.get("project_id") or not service_account.get(
            "private_key"
        ):
            raise Exception(
                "Invalid service account JSON. Required fields are missing."
            )

        return {
            "dataset_id": dataset_name,
            "project": service_account.get("project_id"),
            "credentials": service_account,
        }
    except (ValueError, TypeError) as e:
        print("Invalid service account JSON.", e)
    return connection_string


def connect_to_bigquery(config, using_connection_string):
    if using_connection_string:
        credentials = service_account.Credentials.from_service_account_info(
            config["credentials"]
        )
    else:
        credentials = service_account.Credentials.from_service_account_file(
            config["service_account_file_path"]
        )
    return bigquery.Client(project=config["project"], credentials=credentials)


def run_query_big_query(query, connection):
    query_job = connection.query(query)
    result = query_job.result()
    rows = [dict(row) for row in result]
    fields = [
        {
            "name": field.name,
            "dataTypeID": convert_bigquery_to_postgres(field.field_type),
        }
        for field in result.schema
    ]
    # TODO CONVERT to postgres types

    return {"rows": rows, "fields": fields}


def get_tables_by_schema_big_query(connection, schema_names):
    all_table = []
    for schema_name in schema_names:
        dataset_ref = connection.dataset(schema_name)
        tables = connection.list_tables(dataset_ref)
        for table in tables:
            cur_table = {}
            cur_table["table_name"] = table.table_id
            cur_table["schema_name"] = schema_name
            all_table.append(cur_table)
    return all_table


def get_schema_column_info_big_query(connection, schema_name, table_names):
    all_columns = []
    for table_name in table_names:
        table_ref = connection.dataset(table_name["schema_name"]).table(
            table_name["table_name"]
        )
        table = connection.get_table(table_ref)
        columns = []
        for field in table.schema:
            columns.append(
                {
                    "columnName": field.name,
                    "displayName": field.name,
                    "dataTypeId": convert_bigquery_to_postgres(field.field_type),
                    "fieldType": field.field_type,
                }
            )
        all_columns.append(
            {
                "tableName": table_name["schema_name"] + "." + table_name["table_name"],
                "displayName": table_name["schema_name"]
                + "."
                + table_name["table_name"],
                "columns": columns,
            }
        )
    return all_columns


def infer_schema_big_query(elem):
    # compare elem with regex
    if isinstance(elem, list):
        return 23
    if isinstance(elem, object):
        if re.match(r"/^\d{4}-\d{2}-\d{2}$/", elem.get("value")):
            return 1082
        elif re.match(r"/^\d{2}\/\d{2}\/\d{2,4}$/", elem.get("value")):
            return 1082
        elif re.match(
            r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$/", elem.get("value")
        ):
            return 1184
        elif re.match(
            r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?$/", elem.get("value")
        ):
            return 1114
        elif re.match(r"/^\d{2}:\d{2}:\d{2}$/", elem.get("value")):
            return 1083
    if isinstance(elem, str):
        if re.match(r"/^\d{4}-\d{2}-\d{2}$/", elem):
            return 1082
        elif re.match(r"/^\d{2}\/\d{2}\/\d{2,4}$/", elem):
            return 1082
        elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$/", elem):
            return 1184
        elif re.match(r"/^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?$/", elem):
            return 1114
        elif re.match(r"/^\d{2}:\d{2}:\d{2}$/", elem):
            return 1083
        else:
            return 1043
    return 1043


def convert_bigquery_to_postgres(value):
    type_to_oid = {
        "VARCHAR": 1043,
        "INTEGER": 23,
        "FLOAT": 700,
        "TIMESTAMP": 1114,
        "DATE": 1082,
        "BOOL": 16,
    }
    return type_to_oid.get(value.upper()) or 1043



import json
import logging
import asyncio
from datetime import date, datetime, time
from decimal import Decimal
from uuid import UUID
import redis
from quillsql.error import PgQueryError
from quillsql.db.db_helper import (
    connect_to_db,
    run_query_by_db,
    disconnect_from_db,
    create_async_connection,
    run_query_by_db_async,
    disconnect_from_db_async,
)

# The TTL for new cache entries (default: 1h)
DEFAULT_CACHE_TTL = 24 * 60 * 60
DEFAULT_REDIS_TIMEOUT_SECONDS = 3
logger = logging.getLogger(__name__)

def _cache_json_default(value):
    # Match JS JSON behavior for common DB scalar types.
    if isinstance(value, Decimal):
        return str(value)
    if isinstance(value, (datetime, date, time)):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    if isinstance(value, (bytes, bytearray)):
        try:
            return value.decode("utf-8")
        except Exception:
            return value.hex()
    return str(value)


class CachedConnection:
    def __init__(self, database_type, config, cache_config, using_connection_string):
        self.database_type = database_type
        self.using_connection_string = using_connection_string
        self.config = config
        self.connection = None
        self.async_connection = None
        self._async_open_lock = asyncio.Lock()
        if self._is_postgres():
            self.async_connection = create_async_connection(
                database_type, config, using_connection_string
            )
        else:
            self.connection = connect_to_db(database_type, config, using_connection_string)
        self.cache = self.get_cache(cache_config)
        self.ttl = cache_config and cache_config.get("ttl") or DEFAULT_CACHE_TTL

    def _is_postgres(self):
        return self.database_type and self.database_type.lower() in (
            "postgresql",
            "postgres",
        )

    def get_cache(self, cache_config):
        cache_type = cache_config and cache_config.get("cache_type")
        if cache_type not in ("redis", "rediss"):
            return None
        try:
            timeout = cache_config.get(
                "timeout_seconds", DEFAULT_REDIS_TIMEOUT_SECONDS
            )
            client = redis.Redis(
                host=cache_config.get("host", "localhost"),
                port=cache_config.get("port", 6379),
                username=cache_config.get("username", "default"),
                password=cache_config.get("password"),
                socket_connect_timeout=timeout,
                socket_timeout=timeout,
                retry_on_timeout=True,
            )
            # Force auth/connect validation up-front.
            client.ping()
            return client
        except Exception as err:
            logger.warning(
                "Redis cache unavailable; continuing without cache: %s", str(err)
            )
        return None

    def _is_connection_closed(self):
        return self.connection is None or getattr(self.connection, "closed", True)

    def _open_connection(self):
        if self._is_postgres():
            raise RuntimeError(
                "Synchronous Postgres queries are not supported. "
                "Use async query methods."
            )
        self.connection = connect_to_db(
            self.database_type, self.config, self.using_connection_string
        )
        return self.connection

    def _is_async_connection_closed(self):
        return self.async_connection is None or getattr(
            self.async_connection, "closed", True
        )

    async def _open_connection_async(self):
        if self.async_connection is None:
            self.async_connection = create_async_connection(
                self.database_type, self.config, self.using_connection_string
            )
        if self._is_postgres():
            await self.async_connection.open(wait=False)
        return self.async_connection

    async def ensure_open(self):
        if not self._is_async_connection_closed():
            return self.async_connection
        async with self._async_open_lock:
            if self._is_async_connection_closed():
                await self._open_connection_async()
        return self.async_connection

    def _close_connection(self):
        if self.connection is None:
            return
        try:
            disconnect_from_db(self.database_type, self.connection)
        except Exception:
            pass
        finally:
            self.connection = None

    def close(self):
        self._close_connection()

    async def _close_connection_async(self):
        if self.async_connection is None:
            return
        try:
            await disconnect_from_db_async(self.database_type, self.async_connection)
        except Exception:
            pass
        finally:
            self.async_connection = None

    async def close_async(self):
        await self._close_connection_async()
        self._close_connection()

    def exec(self, sql):
        try:
            if self._is_connection_closed():
                self._open_connection()
            return run_query_by_db(self.database_type, sql, self.connection)
        except Exception as err:
            diag = getattr(err, "diag", None)
            position = getattr(diag, "statement_position", None)
            raise PgQueryError(err, sql, position)

    async def exec_async(self, sql):
        try:
            await self.ensure_open()
            return await run_query_by_db_async(
                self.database_type, sql, self.async_connection
            )
        except Exception as err:
            diag = getattr(err, "diag", None)
            position = getattr(diag, "statement_position", None)
            raise PgQueryError(err, sql, position)

    def query(self, sql, overwrite_cache=False):
        if not self.cache:
            return self.exec(sql)

        key = sql
        if not overwrite_cache:
            try:
                cached_result = self.cache.get(key)
            except Exception as err:
                logger.warning("Redis cache read failed; skipping cache: %s", str(err))
                cached_result = None
            if cached_result:
                try:
                    cached = json.loads(cached_result)
                except Exception as err:
                    logger.warning(
                        "Redis cached value parse failed; bypassing cache entry: %s",
                        str(err),
                    )
                    cached = None
                if isinstance(cached, dict):
                    return {
                        "rows": cached.get("rows", []),
                        "fields": cached.get("fields", []),
                    }
                # Backwards-compat: support legacy cache values containing only rows.
                if cached is not None:
                    return {"rows": cached, "fields": []}

        new_result = self.exec(sql)
        try:
            new_result_string = json.dumps(new_result, default=_cache_json_default)
            # Match Node query-cache behavior: no EX ttl in this path.
            self.cache.set(key, new_result_string)
        except Exception as err:
            logger.warning("Redis cache write failed; skipping cache: %s", str(err))
        return {
            "rows": new_result.get("rows", []),
            "fields": new_result.get("fields", []),
        }

    async def query_async(self, sql, overwrite_cache=False):
        await self.ensure_open()
        if not self.cache:
            return await self.exec_async(sql)

        key = sql
        if not overwrite_cache:
            try:
                cached_result = self.cache.get(key)
            except Exception as err:
                logger.warning("Redis cache read failed; skipping cache: %s", str(err))
                cached_result = None
            if cached_result:
                try:
                    cached = json.loads(cached_result)
                except Exception as err:
                    logger.warning(
                        "Redis cached value parse failed; bypassing cache entry: %s",
                        str(err),
                    )
                    cached = None
                if isinstance(cached, dict):
                    return {
                        "rows": cached.get("rows", []),
                        "fields": cached.get("fields", []),
                    }
                if cached is not None:
                    return {"rows": cached, "fields": []}

        new_result = await self.exec_async(sql)
        try:
            new_result_string = json.dumps(new_result, default=_cache_json_default)
            self.cache.set(key, new_result_string)
        except Exception as err:
            logger.warning("Redis cache write failed; skipping cache: %s", str(err))
        return {
            "rows": new_result.get("rows", []),
            "fields": new_result.get("fields", []),
        }
